from __future__ import print_function, unicode_literals
import io
from collections import namedtuple
from unittest import mock
from twisted.internet import endpoints, reactor
from twisted.trial import unittest
from .._hints import (endpoint_from_hint_obj, parse_hint_argv, parse_tcp_v1_hint,
                      describe_hint_obj, parse_hint, encode_hint,
                      DirectTCPV1Hint, TorTCPV1Hint, RelayV1Hint)

UnknownHint = namedtuple("UnknownHint", ["stuff"])


class Hints(unittest.TestCase):
    def test_endpoint_from_hint_obj(self):
        def efho(hint, tor=None):
            return endpoint_from_hint_obj(hint, tor, reactor)
        self.assertIsInstance(efho(DirectTCPV1Hint("host", 1234, 0.0)),
                              endpoints.HostnameEndpoint)
        self.assertEqual(efho("unknown:stuff:yowza:pivlor"), None)

        # tor=None
        self.assertEqual(efho(TorTCPV1Hint("host", "port", 0)), None)
        self.assertEqual(efho(UnknownHint("foo")), None)

        tor = mock.Mock()

        def tor_ep(hostname, port):
            if hostname == "non-public":
                raise ValueError
            return ("tor_ep", hostname, port)
        tor.stream_via = mock.Mock(side_effect=tor_ep)

        self.assertEqual(efho(DirectTCPV1Hint("host", 1234, 0.0), tor),
                         ("tor_ep", "host", 1234))
        self.assertEqual(efho(TorTCPV1Hint("host2.onion", 1234, 0.0), tor),
                         ("tor_ep", "host2.onion", 1234))
        self.assertEqual(efho(DirectTCPV1Hint("non-public", 1234, 0.0), tor), None)

        self.assertEqual(efho(UnknownHint("foo"), tor), None)

    def test_comparable(self):
        h1 = DirectTCPV1Hint("hostname", "port1", 0.0)
        h1b = DirectTCPV1Hint("hostname", "port1", 0.0)
        h2 = DirectTCPV1Hint("hostname", "port2", 0.0)
        r1 = RelayV1Hint(tuple(sorted([h1, h2])))
        r2 = RelayV1Hint(tuple(sorted([h2, h1])))
        r3 = RelayV1Hint(tuple(sorted([h1b, h2])))
        self.assertEqual(r1, r2)
        self.assertEqual(r2, r3)
        self.assertEqual(len(set([r1, r2, r3])), 1)

    def test_parse_tcp_v1_hint(self):
        p = parse_tcp_v1_hint
        self.assertEqual(p({"type": "unknown"}), None)
        h = p({"type": "direct-tcp-v1", "hostname": "foo", "port": 1234})
        self.assertEqual(h, DirectTCPV1Hint("foo", 1234, 0.0))
        h = p({
            "type": "direct-tcp-v1",
            "hostname": "foo",
            "port": 1234,
            "priority": 2.5
        })
        self.assertEqual(h, DirectTCPV1Hint("foo", 1234, 2.5))
        h = p({"type": "tor-tcp-v1", "hostname": "foo", "port": 1234})
        self.assertEqual(h, TorTCPV1Hint("foo", 1234, 0.0))
        h = p({
            "type": "tor-tcp-v1",
            "hostname": "foo",
            "port": 1234,
            "priority": 2.5
        })
        self.assertEqual(h, TorTCPV1Hint("foo", 1234, 2.5))
        self.assertEqual(p({
            "type": "direct-tcp-v1"
        }), None)  # missing hostname
        self.assertEqual(p({
            "type": "direct-tcp-v1",
            "hostname": 12
        }), None)  # invalid hostname
        self.assertEqual(
            p({
                "type": "direct-tcp-v1",
                "hostname": "foo"
            }), None)  # missing port
        self.assertEqual(
            p({
                "type": "direct-tcp-v1",
                "hostname": "foo",
                "port": "not a number"
            }), None)  # invalid port

    def test_parse_hint(self):
        p = parse_hint
        self.assertEqual(p({"type": "direct-tcp-v1",
                            "hostname": "foo",
                            "port": 12}),
                         DirectTCPV1Hint("foo", 12, 0.0))
        self.assertEqual(p({"type": "relay-v1",
                            "hints": [
                                {"type": "direct-tcp-v1",
                                 "hostname": "foo",
                                 "port": 12},
                                {"type": "unrecognized"},
                                {"type": "direct-tcp-v1",
                                 "hostname": "bar",
                                 "port": 13}]}),
                         RelayV1Hint([DirectTCPV1Hint("foo", 12, 0.0),
                                      DirectTCPV1Hint("bar", 13, 0.0)]))

    def test_parse_hint_argv(self):
        def p(hint):
            stderr = io.StringIO()
            value = parse_hint_argv(hint, stderr=stderr)
            return value, stderr.getvalue()

        h, stderr = p("tcp:host:1234")
        self.assertEqual(h, DirectTCPV1Hint("host", 1234, 0.0))
        self.assertEqual(stderr, "")

        h, stderr = p("tcp:host:1234:priority=2.6")
        self.assertEqual(h, DirectTCPV1Hint("host", 1234, 2.6))
        self.assertEqual(stderr, "")

        h, stderr = p("tcp:host:1234:unknown=stuff")
        self.assertEqual(h, DirectTCPV1Hint("host", 1234, 0.0))
        self.assertEqual(stderr, "")

        h, stderr = p("$!@#^")
        self.assertEqual(h, None)
        self.assertEqual(stderr, "unparseable hint '$!@#^'\n")

        h, stderr = p("unknown:stuff")
        self.assertEqual(h, None)
        self.assertEqual(stderr,
                         "unknown hint type 'unknown' in 'unknown:stuff'\n")

        h, stderr = p("tcp:just-a-hostname")
        self.assertEqual(h, None)
        self.assertEqual(
            stderr,
            "unparseable TCP hint (need more colons) 'tcp:just-a-hostname'\n")

        h, stderr = p("tcp:host:number")
        self.assertEqual(h, None)
        self.assertEqual(stderr,
                         "non-numeric port in TCP hint 'tcp:host:number'\n")

        h, stderr = p("tcp:host:1234:priority=bad")
        self.assertEqual(h, None)
        self.assertEqual(
            stderr,
            "non-float priority= in TCP hint 'tcp:host:1234:priority=bad'\n")

        h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]")
        self.assertEqual(h, None)
        self.assertEqual(
            stderr, "non-numeric port in TCP hint 'tcp:[2001:0db8:85a3::8a2e:0370:7334]'\n")

        h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]:1234")
        self.assertEqual(h, DirectTCPV1Hint(
            "2001:0db8:85a3::8a2e:0370:7334", 1234, 0.0))
        self.assertEqual(stderr, "")

        h, stderr = p("tcp:[2001:0db8:85a3::8a2e:0370:7334]:1234:priority=2.6")
        self.assertEqual(h, DirectTCPV1Hint(
            "2001:0db8:85a3::8a2e:0370:7334", 1234, 2.6))
        self.assertEqual(stderr, "")

        h, stderr = p("tcp:[abc::xyz]:1234")
        self.assertEqual(h, None)
        self.assertEqual(
            stderr, "invalid IPv6 address in TCP hint 'tcp:[abc::xyz]:1234'\n")

    def test_describe_hint_obj(self):
        d = describe_hint_obj
        self.assertEqual(d(DirectTCPV1Hint("host", 1234, 0.0), False, False),
                         "->tcp:host:1234")
        self.assertEqual(d(DirectTCPV1Hint("host", 1234, 0.0), True, False),
                         "->relay:tcp:host:1234")
        self.assertEqual(d(DirectTCPV1Hint("host", 1234, 0.0), False, True),
                         "tor->tcp:host:1234")
        self.assertEqual(d(DirectTCPV1Hint("host", 1234, 0.0), True, True),
                         "tor->relay:tcp:host:1234")
        self.assertEqual(d(TorTCPV1Hint("host", 1234, 0.0), False, False),
                         "->tor:host:1234")
        self.assertEqual(d(UnknownHint("stuff"), False, False),
                         "->%s" % str(UnknownHint("stuff")))

    def test_encode_hint(self):
        e = encode_hint
        self.assertEqual(e(DirectTCPV1Hint("host", 1234, 1.0)),
                         {"type": "direct-tcp-v1",
                          "priority": 1.0,
                          "hostname": "host",
                          "port": 1234})
        self.assertEqual(e(RelayV1Hint([DirectTCPV1Hint("foo", 12, 0.0),
                                        DirectTCPV1Hint("bar", 13, 0.0)])),
                         {"type": "relay-v1",
                          "hints": [
                              {"type": "direct-tcp-v1",
                               "hostname": "foo",
                               "port": 12,
                               "priority": 0.0},
                              {"type": "direct-tcp-v1",
                               "hostname": "bar",
                               "port": 13,
                               "priority": 0.0},
                          ]})
        self.assertEqual(e(TorTCPV1Hint("host", 1234, 1.0)),
                         {"type": "tor-tcp-v1",
                          "priority": 1.0,
                          "hostname": "host",
                          "port": 1234})
        e = self.assertRaises(ValueError, e, "not a Hint")
        self.assertIn("unknown hint type", str(e))
        self.assertIn("not a Hint", str(e))



import io
import os

from twisted.trial import unittest

from unittest import mock

from ..cli import cmd_ssh

OTHERS = ["config", "config~", "known_hosts", "known_hosts~"]


class FindPubkey(unittest.TestCase):
    def test_find_one(self):
        files = OTHERS + ["id_rsa.pub", "id_rsa"]
        pubkey_data = u"ssh-rsa AAAAkeystuff email@host\n"
        pubkey_file = io.StringIO(pubkey_data)
        with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
            with mock.patch("os.listdir", return_value=files) as ld:
                with mock.patch(
                        "wormhole.cli.cmd_ssh.open", return_value=pubkey_file):
                    res = cmd_ssh.find_public_key()
        self.assertEqual(ld.mock_calls,
                         [mock.call(os.path.expanduser("~/.ssh/"))])
        self.assertEqual(len(res), 3, res)
        kind, keyid, pubkey = res
        self.assertEqual(kind, "ssh-rsa")
        self.assertEqual(keyid, "email@host")
        self.assertEqual(pubkey, pubkey_data)

    def test_find_none(self):
        files = OTHERS  # no pubkey
        with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
            with mock.patch("os.listdir", return_value=files):
                e = self.assertRaises(cmd_ssh.PubkeyError,
                                      cmd_ssh.find_public_key)
        dot_ssh = os.path.expanduser("~/.ssh/")
        self.assertEqual(str(e), "No public keys in '{}'".format(dot_ssh))

    def test_bad_hint(self):
        with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=False):
            e = self.assertRaises(
                cmd_ssh.PubkeyError,
                cmd_ssh.find_public_key,
                hint="bogus/path")
        self.assertEqual(str(e), "Can't find 'bogus/path'")

    def test_find_multiple(self):
        files = OTHERS + ["id_rsa.pub", "id_rsa", "id_dsa.pub", "id_dsa"]
        pubkey_data = u"ssh-rsa AAAAkeystuff email@host\n"
        pubkey_file = io.StringIO(pubkey_data)
        with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
            with mock.patch("os.listdir", return_value=files):
                responses = iter(["frog", "NaN", "-1", "0"])
                with mock.patch(
                        "click.prompt", side_effect=lambda p: next(responses)):
                    with mock.patch(
                            "wormhole.cli.cmd_ssh.open",
                            return_value=pubkey_file):
                        res = cmd_ssh.find_public_key()
        self.assertEqual(len(res), 3, res)
        kind, keyid, pubkey = res
        self.assertEqual(kind, "ssh-rsa")
        self.assertEqual(keyid, "email@host")
        self.assertEqual(pubkey, pubkey_data)

    def test_comment_with_spaces(self):
        files = OTHERS + ["id_ed25519.pub", "id_ed25519"]
        pubkey_data = u"ssh-ed25519 AAAAkeystuff comment with spaces"
        pubkey_file = io.StringIO(pubkey_data)

        with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
            with mock.patch("os.listdir", return_value=files) as ld:
                with mock.patch(
                        "wormhole.cli.cmd_ssh.open", return_value=pubkey_file):
                    res = cmd_ssh.find_public_key()
        self.assertEqual(ld.mock_calls,
                         [mock.call(os.path.expanduser("~/.ssh/"))])
        self.assertEqual(len(res), 3, res)
        kind, keyid, pubkey = res
        self.assertEqual(kind, "ssh-ed25519")
        self.assertEqual(keyid, "comment with spaces")
        self.assertEqual(pubkey, pubkey_data)



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import directlyProvides
from twisted.trial import unittest
from twisted.internet.interfaces import ITransport, IHalfCloseableProtocol
from twisted.internet.error import ConnectionDone
from ..._dilation.subchannel import (Once, SubChannel,
                                     _WormholeAddress, _SubchannelAddress,
                                     AlreadyClosedError,
                                     NormalCloseUsedOnHalfCloseable)
from .common import mock_manager


def make_sc(set_protocol=True, half_closeable=False):
    scid = 4
    hostaddr = _WormholeAddress()
    peeraddr = _SubchannelAddress(scid)
    m = mock_manager()
    sc = SubChannel(scid, m, hostaddr, peeraddr)
    p = mock.Mock()
    if half_closeable:
        directlyProvides(p, IHalfCloseableProtocol)
    if set_protocol:
        sc._set_protocol(p)
    return sc, m, scid, hostaddr, peeraddr, p


class SubChannelAPI(unittest.TestCase):
    def test_once(self):
        o = Once(ValueError)
        o()
        with self.assertRaises(ValueError):
            o()

    def test_create(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()
        self.assert_(ITransport.providedBy(sc))
        self.assertEqual(m.mock_calls, [])
        self.assertIdentical(sc.getHost(), hostaddr)
        self.assertIdentical(sc.getPeer(), peeraddr)

    def test_write(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        sc.write(b"data")
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"data")])
        m.mock_calls[:] = []
        sc.writeSequence([b"more", b"data"])
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"moredata")])

    def test_write_when_closing(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        sc.loseConnection()
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid)])
        m.mock_calls[:] = []

        with self.assertRaises(AlreadyClosedError) as e:
            sc.write(b"data")
        self.assertEqual(str(e.exception),
                         "write not allowed on closed subchannel")

    def test_local_close(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        sc.loseConnection()
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid)])
        m.mock_calls[:] = []

        # late arriving data is still delivered
        sc.remote_data(b"late")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"late")])
        p.mock_calls[:] = []

        sc.remote_close()
        self.assert_connectionDone(p.mock_calls)

    def test_local_double_close(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        sc.loseConnection()
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid)])
        m.mock_calls[:] = []

        with self.assertRaises(AlreadyClosedError) as e:
            sc.loseConnection()
        self.assertEqual(str(e.exception),
                         "loseConnection not allowed on closed subchannel")

    def assert_connectionDone(self, mock_calls):
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][0], "connectionLost")
        self.assertEqual(len(mock_calls[0][1]), 1)
        self.assertIsInstance(mock_calls[0][1][0], ConnectionDone)

    def test_remote_close(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()
        sc.remote_close()
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid),
                                        mock.call.subchannel_closed(scid, sc)])
        self.assert_connectionDone(p.mock_calls)

    def test_data(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()
        sc.remote_data(b"data")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"data")])
        p.mock_calls[:] = []
        sc.remote_data(b"not")
        sc.remote_data(b"coalesced")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"not"),
                                        mock.call.dataReceived(b"coalesced"),
                                        ])

    def test_data_before_open(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc(set_protocol=False)
        sc.remote_data(b"data1")
        sc.remote_data(b"data2")
        self.assertEqual(p.mock_calls, [])
        sc._set_protocol(p)
        sc._deliver_queued_data()
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"data1"),
                                        mock.call.dataReceived(b"data2")])
        p.mock_calls[:] = []
        sc.remote_data(b"more")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"more")])

    def test_close_before_open(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc(set_protocol=False)
        sc.remote_close()
        self.assertEqual(p.mock_calls, [])
        sc._set_protocol(p)
        sc._deliver_queued_data()
        self.assert_connectionDone(p.mock_calls)

    def test_producer(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        sc.pauseProducing()
        self.assertEqual(m.mock_calls, [mock.call.subchannel_pauseProducing(sc)])
        m.mock_calls[:] = []
        sc.resumeProducing()
        self.assertEqual(m.mock_calls, [mock.call.subchannel_resumeProducing(sc)])
        m.mock_calls[:] = []
        sc.stopProducing()
        self.assertEqual(m.mock_calls, [mock.call.subchannel_stopProducing(sc)])
        m.mock_calls[:] = []

    def test_consumer(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc()

        # TODO: more, once this is implemented
        sc.registerProducer(None, True)
        sc.unregisterProducer()


class HalfCloseable(unittest.TestCase):

    def test_create(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)
        self.assert_(ITransport.providedBy(sc))
        self.assertEqual(m.mock_calls, [])
        self.assertIdentical(sc.getHost(), hostaddr)
        self.assertIdentical(sc.getPeer(), peeraddr)

    def test_local_close(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)

        sc.write(b"data")
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"data")])
        m.mock_calls[:] = []
        sc.writeSequence([b"more", b"data"])
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"moredata")])
        m.mock_calls[:] = []

        sc.remote_data(b"inbound1")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"inbound1")])
        p.mock_calls[:] = []

        with self.assertRaises(NormalCloseUsedOnHalfCloseable) as e:
            sc.loseConnection()  # TODO: maybe this shouldn't be an error

        # after a local close, we can't write anymore, but we can still
        # receive data
        sc.loseWriteConnection()  # TODO or loseConnection?
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid)])
        m.mock_calls[:] = []
        self.assertEqual(p.mock_calls, [mock.call.writeConnectionLost()])
        p.mock_calls[:] = []

        with self.assertRaises(AlreadyClosedError) as e:
            sc.write(b"data")
        self.assertEqual(str(e.exception),
                         "write not allowed on closed subchannel")

        with self.assertRaises(AlreadyClosedError) as e:
            sc.loseWriteConnection()
        self.assertEqual(str(e.exception),
                         "loseConnection not allowed on closed subchannel")

        with self.assertRaises(NormalCloseUsedOnHalfCloseable) as e:
            sc.loseConnection()  # TODO: maybe expect AlreadyClosedError

        sc.remote_data(b"inbound2")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"inbound2")])
        p.mock_calls[:] = []

        # the remote end will finally shut down the connection
        sc.remote_close()
        self.assertEqual(m.mock_calls, [mock.call.subchannel_closed(scid, sc)])
        self.assertEqual(p.mock_calls, [mock.call.readConnectionLost()])

    def test_remote_close(self):
        sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)

        sc.write(b"data")
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"data")])
        m.mock_calls[:] = []

        sc.remote_data(b"inbound1")
        self.assertEqual(p.mock_calls, [mock.call.dataReceived(b"inbound1")])
        p.mock_calls[:] = []

        # after a remote close, we can still write data
        sc.remote_close()
        self.assertEqual(m.mock_calls, [])
        self.assertEqual(p.mock_calls, [mock.call.readConnectionLost()])
        p.mock_calls[:] = []

        sc.write(b"out2")
        self.assertEqual(m.mock_calls, [mock.call.send_data(scid, b"out2")])
        m.mock_calls[:] = []

        # and a local close will shutdown the connection
        sc.loseWriteConnection()
        self.assertEqual(m.mock_calls, [mock.call.send_close(scid),
                                        mock.call.subchannel_closed(scid, sc)])
        self.assertEqual(p.mock_calls, [mock.call.writeConnectionLost()])



from collections import deque
from attr import attrs, attrib
from attr.validators import instance_of, provides
from zope.interface import implementer
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.interfaces import (ITransport, IProducer, IConsumer,
                                         IAddress, IListeningPort,
                                         IHalfCloseableProtocol,
                                         IStreamClientEndpoint,
                                         IStreamServerEndpoint)
from twisted.internet.error import ConnectionDone
from automat import MethodicalMachine
from .._interfaces import ISubChannel, IDilationManager
from ..observer import OneShotObserver

# each subchannel frame (the data passed into transport.write(data)) gets a
# 9-byte header prefix (type, subchannel id, and sequence number), then gets
# encrypted (adding a 16-byte authentication tag). The result is transmitted
# with a 4-byte length prefix (which only covers the padded message, not the
# length prefix itself), so the padded message must be less than 2**32 bytes
# long.
MAX_FRAME_LENGTH = 2**32 - 1 - 9 - 16


@attrs
class Once(object):
    _errtype = attrib()

    def __attrs_post_init__(self):
        self._called = False

    def __call__(self):
        if self._called:
            raise self._errtype()
        self._called = True


class SingleUseEndpointError(Exception):
    pass

# created in the (OPEN) state, by either:
#  * receipt of an OPEN message
#  * or local client_endpoint.connect()
# then transitions are:
# (OPEN) rx DATA: deliver .dataReceived(), -> (OPEN)
# (OPEN) rx CLOSE: deliver .connectionLost(), send CLOSE, -> (CLOSED)
# (OPEN) local .write(): send DATA, -> (OPEN)
# (OPEN) local .loseConnection(): send CLOSE, -> (CLOSING)
# (CLOSING) local .write(): error
# (CLOSING) local .loseConnection(): error
# (CLOSING) rx DATA: deliver .dataReceived(), -> (CLOSING)
# (CLOSING) rx CLOSE: deliver .connectionLost(), -> (CLOSED)
# object is deleted upon transition to (CLOSED)


class AlreadyClosedError(Exception):
    pass


class NormalCloseUsedOnHalfCloseable(Exception):
    pass


class HalfCloseUsedOnNonHalfCloseable(Exception):
    pass


@implementer(IAddress)
class _WormholeAddress(object):
    pass


@implementer(IAddress)
@attrs
class _SubchannelAddress(object):
    _scid = attrib(validator=instance_of(int))


@attrs(eq=False)
@implementer(ITransport)
@implementer(IProducer)
@implementer(IConsumer)
@implementer(ISubChannel)
class SubChannel(object):
    _scid = attrib(validator=instance_of(int))
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib(validator=instance_of(_WormholeAddress))
    _peer_addr = attrib(validator=instance_of(_SubchannelAddress))

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self,
                        f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        # self._mailbox = None
        # self._pending_outbound = {}
        # self._processed = set()
        self._protocol = None
        self._pending_remote_data = []
        self._pending_remote_close = False

    @m.state(initial=True)
    def unconnected(self):
        pass  # pragma: no cover

    # once we get the IProtocol, it's either a IHalfCloseableProtocol, or it
    # can only be fully closed
    @m.state()
    def open_half(self):
        pass  # pragma: no cover

    @m.state()
    def read_closed():
        pass  # pragma: no cover

    @m.state()
    def write_closed():
        pass  # pragma: no cover

    @m.state()
    def open_full(self):
        pass  # pragma: no cover

    @m.state()
    def closing():
        pass  # pragma: no cover

    @m.state()
    def closed():
        pass  # pragma: no cover

    @m.input()
    def connect_protocol_half(self):
        pass

    @m.input()
    def connect_protocol_full(self):
        pass

    @m.input()
    def remote_data(self, data):
        pass

    @m.input()
    def remote_close(self):
        pass

    @m.input()
    def local_data(self, data):
        pass

    @m.input()
    def local_close(self):
        pass

    @m.output()
    def queue_remote_data(self, data):
        self._pending_remote_data.append(data)

    @m.output()
    def queue_remote_close(self):
        self._pending_remote_close = True

    @m.output()
    def send_data(self, data):
        self._manager.send_data(self._scid, data)

    @m.output()
    def send_close(self):
        self._manager.send_close(self._scid)

    @m.output()
    def signal_dataReceived(self, data):
        assert self._protocol
        self._protocol.dataReceived(data)

    @m.output()
    def signal_readConnectionLost(self):
        IHalfCloseableProtocol(self._protocol).readConnectionLost()

    @m.output()
    def signal_writeConnectionLost(self):
        IHalfCloseableProtocol(self._protocol).writeConnectionLost()

    @m.output()
    def signal_connectionLost(self):
        assert self._protocol
        self._protocol.connectionLost(ConnectionDone())

    @m.output()
    def close_subchannel(self):
        self._manager.subchannel_closed(self._scid, self)
        # we're deleted momentarily

    @m.output()
    def error_closed_write(self, data):
        raise AlreadyClosedError("write not allowed on closed subchannel")

    @m.output()
    def error_closed_close(self):
        raise AlreadyClosedError(
            "loseConnection not allowed on closed subchannel")

    # stuff that arrives before we have a protocol connected
    unconnected.upon(remote_data, enter=unconnected, outputs=[queue_remote_data])
    unconnected.upon(remote_close, enter=unconnected, outputs=[queue_remote_close])

    # IHalfCloseableProtocol flow
    unconnected.upon(connect_protocol_half, enter=open_half, outputs=[])
    open_half.upon(remote_data, enter=open_half, outputs=[signal_dataReceived])
    open_half.upon(local_data, enter=open_half, outputs=[send_data])
    # remote closes first
    open_half.upon(remote_close, enter=read_closed, outputs=[signal_readConnectionLost])
    read_closed.upon(local_data, enter=read_closed, outputs=[send_data])
    read_closed.upon(local_close, enter=closed, outputs=[send_close,
                                                         close_subchannel,
                                                         # TODO: eventual-signal this?
                                                         signal_writeConnectionLost,
                                                         ])
    # local closes first
    open_half.upon(local_close, enter=write_closed, outputs=[signal_writeConnectionLost,
                                                             send_close])
    write_closed.upon(local_data, enter=write_closed, outputs=[error_closed_write])
    write_closed.upon(remote_data, enter=write_closed, outputs=[signal_dataReceived])
    write_closed.upon(remote_close, enter=closed, outputs=[close_subchannel,
                                                           signal_readConnectionLost,
                                                           ])
    # error cases
    write_closed.upon(local_close, enter=write_closed, outputs=[error_closed_close])

    # fully-closeable-only flow
    unconnected.upon(connect_protocol_full, enter=open_full, outputs=[])
    open_full.upon(remote_data, enter=open_full, outputs=[signal_dataReceived])
    open_full.upon(local_data, enter=open_full, outputs=[send_data])
    open_full.upon(remote_close, enter=closed, outputs=[send_close,
                                                        close_subchannel,
                                                        signal_connectionLost])
    open_full.upon(local_close, enter=closing, outputs=[send_close])
    closing.upon(remote_data, enter=closing, outputs=[signal_dataReceived])
    closing.upon(remote_close, enter=closed, outputs=[close_subchannel,
                                                      signal_connectionLost])
    # error cases
    # we won't ever see an OPEN, since L4 will log+ignore those for us
    closing.upon(local_data, enter=closing, outputs=[error_closed_write])
    closing.upon(local_close, enter=closing, outputs=[error_closed_close])
    # the CLOSED state shouldn't ever see messages, since we'll be deleted
    # (but a local user should be able to call "close" without having
    # to know what state we're in)
    closed.upon(local_close, enter=closed, outputs=[])

    # our endpoints use these

    def _set_protocol(self, protocol):
        assert not self._protocol
        self._protocol = protocol
        if IHalfCloseableProtocol.providedBy(protocol):
            self.connect_protocol_half()
        else:
            # move from UNCONNECTED to OPEN
            self.connect_protocol_full()

    def _deliver_queued_data(self):
        for data in self._pending_remote_data:
            self.remote_data(data)
        del self._pending_remote_data
        if self._pending_remote_close:
            self.remote_close()
            del self._pending_remote_close

    # ITransport
    def write(self, data):
        assert isinstance(data, type(b""))
        assert len(data) <= MAX_FRAME_LENGTH
        self.local_data(data)

    def writeSequence(self, iovec):
        self.write(b"".join(iovec))

    def loseWriteConnection(self):
        if not IHalfCloseableProtocol.providedBy(self._protocol):
            # this is a clear error
            raise HalfCloseUsedOnNonHalfCloseable()
        self.local_close()

    def loseConnection(self):
        # TODO: what happens if an IHalfCloseableProtocol calls normal
        # loseConnection()? I think we need to close the read side too.
        if IHalfCloseableProtocol.providedBy(self._protocol):
            # I don't know is correct, so avoid this for now
            raise NormalCloseUsedOnHalfCloseable()
        self.local_close()

    def getHost(self):
        # we define "host addr" as the overall wormhole
        return self._host_addr

    def getPeer(self):
        # and "peer addr" as the subchannel within that wormhole
        return self._peer_addr

    # IProducer: throttle inbound data (wormhole "up" to local app's Protocol)
    def stopProducing(self):
        self._manager.subchannel_stopProducing(self)

    def pauseProducing(self):
        self._manager.subchannel_pauseProducing(self)

    def resumeProducing(self):
        self._manager.subchannel_resumeProducing(self)

    # IConsumer: allow the wormhole to throttle outbound data (app->wormhole)
    def registerProducer(self, producer, streaming):
        self._manager.subchannel_registerProducer(self, producer, streaming)

    def unregisterProducer(self):
        self._manager.subchannel_unregisterProducer(self)


@implementer(IStreamClientEndpoint)
@attrs
class ControlEndpoint(object):
    _peer_addr = attrib(validator=provides(IAddress))
    _subchannel_zero = attrib(validator=provides(ISubChannel))
    _eventual_queue = attrib(repr=False)
    _used = False

    def __attrs_post_init__(self):
        self._once = Once(SingleUseEndpointError)
        self._wait_for_main_channel = OneShotObserver(self._eventual_queue)

    # from manager

    def _main_channel_ready(self):
        self._wait_for_main_channel.fire(None)

    def _main_channel_failed(self, f):
        self._wait_for_main_channel.error(f)

    @inlineCallbacks
    def connect(self, protocolFactory):
        # return Deferred that fires with IProtocol or Failure(ConnectError)
        self._once()
        yield self._wait_for_main_channel.when_fired()
        p = protocolFactory.buildProtocol(self._peer_addr)
        self._subchannel_zero._set_protocol(p)
        # this sets p.transport and calls p.connectionMade()
        p.makeConnection(self._subchannel_zero)
        self._subchannel_zero._deliver_queued_data()
        returnValue(p)


@implementer(IStreamClientEndpoint)
@attrs
class SubchannelConnectorEndpoint(object):
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib(validator=instance_of(_WormholeAddress))
    _eventual_queue = attrib(repr=False)

    def __attrs_post_init__(self):
        self._connection_deferreds = deque()
        self._wait_for_main_channel = OneShotObserver(self._eventual_queue)

    def _main_channel_ready(self):
        self._wait_for_main_channel.fire(None)

    def _main_channel_failed(self, f):
        self._wait_for_main_channel.error(f)

    @inlineCallbacks
    def connect(self, protocolFactory):
        # return Deferred that fires with IProtocol or Failure(ConnectError)
        yield self._wait_for_main_channel.when_fired()
        scid = self._manager.allocate_subchannel_id()
        self._manager.send_open(scid)
        peer_addr = _SubchannelAddress(scid)
        # ? f.doStart()
        # ? f.startedConnecting(CONNECTOR) # ??
        sc = SubChannel(scid, self._manager, self._host_addr, peer_addr)
        self._manager.subchannel_local_open(scid, sc)
        p = protocolFactory.buildProtocol(peer_addr)
        sc._set_protocol(p)
        p.makeConnection(sc)  # set p.transport = sc and call connectionMade()
        returnValue(p)


@implementer(IStreamServerEndpoint)
@attrs
class SubchannelListenerEndpoint(object):
    _manager = attrib(validator=provides(IDilationManager))
    _host_addr = attrib(validator=provides(IAddress))
    _eventual_queue = attrib(repr=False)

    def __attrs_post_init__(self):
        self._once = Once(SingleUseEndpointError)
        self._factory = None
        self._pending_opens = deque()
        self._wait_for_main_channel = OneShotObserver(self._eventual_queue)

    # from manager (actually Inbound)
    def _got_open(self, t, peer_addr):
        if self._factory:
            self._connect(t, peer_addr)
        else:
            self._pending_opens.append((t, peer_addr))

    def _connect(self, t, peer_addr):
        p = self._factory.buildProtocol(peer_addr)
        t._set_protocol(p)
        p.makeConnection(t)
        t._deliver_queued_data()

    def _main_channel_ready(self):
        self._wait_for_main_channel.fire(None)

    def _main_channel_failed(self, f):
        self._wait_for_main_channel.error(f)

    # IStreamServerEndpoint

    @inlineCallbacks
    def listen(self, protocolFactory):
        self._once()
        yield self._wait_for_main_channel.when_fired()
        self._factory = protocolFactory
        while self._pending_opens:
            (t, peer_addr) = self._pending_opens.popleft()
            self._connect(t, peer_addr)
        lp = SubchannelListeningPort(self._host_addr)
        returnValue(lp)


@implementer(IListeningPort)
@attrs
class SubchannelListeningPort(object):
    _host_addr = attrib(validator=provides(IAddress))

    def startListening(self):
        pass

    def stopListening(self):
        # TODO
        pass

    def getHost(self):
        return self._host_addr



import os
from collections import deque
from collections.abc import Sequence
from attr import attrs, attrib
from attr.validators import provides, instance_of, optional
from automat import MethodicalMachine
from zope.interface import implementer
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import (IStreamClientEndpoint,
                                         IStreamServerEndpoint)
from twisted.python import log, failure
from .._interfaces import IDilator, IDilationManager, ISend, ITerminator
from ..util import dict_to_bytes, bytes_to_dict, bytes_to_hexstr
from ..observer import OneShotObserver
from .._key import derive_key
from .subchannel import (SubChannel, _SubchannelAddress, _WormholeAddress,
                         ControlEndpoint, SubchannelConnectorEndpoint,
                         SubchannelListenerEndpoint)
from .connector import Connector
from .._hints import parse_hint
from .roles import LEADER, FOLLOWER
from .connection import KCM, Ping, Pong, Open, Data, Close, Ack
from .inbound import Inbound
from .outbound import Outbound


# exported to Wormhole() for inclusion in versions message
# note that these are strings, not numbers, to facilitate
# experimentation or non-standard versions; the _order_ of versions in
# "can-dilate" is important!
DILATION_VERSIONS = ["1"]


class OldPeerCannotDilateError(Exception):
    pass


class UnknownDilationMessageType(Exception):
    pass


class ReceivedHintsTooEarly(Exception):
    pass


class UnexpectedKCM(Exception):
    pass


class UnknownMessageType(Exception):
    pass


@attrs
class EndpointRecord(Sequence):
    control = attrib(validator=provides(IStreamClientEndpoint))
    connect = attrib(validator=provides(IStreamClientEndpoint))
    listen = attrib(validator=provides(IStreamServerEndpoint))

    def __len__(self):
        return 3

    def __getitem__(self, n):
        return (self.control, self.connect, self.listen)[n]


def make_side():
    return bytes_to_hexstr(os.urandom(8))


# new scheme:
# * both sides send PLEASE as soon as they have an unverified key and
#    w.dilate has been called,
# * PLEASE includes a dilation-specific "side" (independent of the "side"
#    used by mailbox messages)
# * higher "side" is Leader, lower is Follower
# * PLEASE includes the selection of a version from the "can-dilate" list of versions, requires overlap
#    "1" is current

# * we start dilation after both w.dilate() and receiving VERSION, putting us
#   in WANTING, then we process all previously-queued inbound DILATE-n
#   messages. When PLEASE arrives, we move to CONNECTING
# * HINTS sent after dilation starts
# * only Leader sends RECONNECT, only Follower sends RECONNECTING. This
#    is the only difference between the two sides, and is not enforced
#    by the protocol (i.e. if the Follower sends RECONNECT to the Leader,
#    the Leader will obey, although TODO how confusing will this get?)
# * upon receiving RECONNECT: drop Connector, start new Connector, send
#   RECONNECTING, start sending HINTS
# * upon sending RECONNECT: go into FLUSHING state and ignore all HINTS until
#   RECONNECTING received. The new Connector can be spun up earlier, and it
#   can send HINTS, but it must not be given any HINTS that arrive before
#   RECONNECTING (since they're probably stale)

# * after VERSIONS(KCM) received, we might learn that the other side cannot
#    dilate. w.dilate errbacks at this point

# * maybe signal warning if we stay in a "want" state for too long
# * nobody sends HINTS until they're ready to receive
# * nobody sends HINTS unless they've called w.dilate() and received PLEASE
# * nobody connects to inbound hints unless they've called w.dilate()
# * if leader calls w.dilate() but not follower, leader waits forever in
#   "want" (doesn't send anything)
# * if follower calls w.dilate() but not leader, follower waits forever
#   in "want", leader waits forever in "wanted"

@attrs(eq=False)
@implementer(IDilationManager)
class Manager(object):
    _S = attrib(validator=provides(ISend), repr=False)
    _my_side = attrib(validator=instance_of(type(u"")))
    _transit_relay_location = attrib(validator=optional(instance_of(str)))
    _reactor = attrib(repr=False)
    _eventual_queue = attrib(repr=False)
    _cooperator = attrib(repr=False)
    # TODO: can this validator work when the parameter is optional?
    _no_listen = attrib(validator=instance_of(bool), default=False)

    _dilation_key = None
    _tor = None  # TODO
    _timing = None  # TODO
    _next_subchannel_id = None  # initialized in choose_role
    _dilation_version = None  # initialized in got_wormhole_versions

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._got_versions_d = Deferred()

        self._my_role = None  # determined upon rx_PLEASE
        self._host_addr = _WormholeAddress()

        self._connection = None
        self._made_first_connection = False
        self._stopped = OneShotObserver(self._eventual_queue)
        self._debug_stall_connector = False

        self._next_dilation_phase = 0

        # I kept getting confused about which methods were for inbound data
        # (and thus flow-control methods go "out") and which were for
        # outbound data (with flow-control going "in"), so I split them up
        # into separate pieces.
        self._inbound = Inbound(self, self._host_addr)
        self._outbound = Outbound(self, self._cooperator)  # from us to peer

        # We must open subchannel0 early, since messages may arrive very
        # quickly once the connection is established. This subchannel may or
        # may not ever get revealed to the caller, since the peer might not
        # even be capable of dilation.
        scid0 = 0
        peer_addr0 = _SubchannelAddress(scid0)
        sc0 = SubChannel(scid0, self, self._host_addr, peer_addr0)
        self._inbound.set_subchannel_zero(scid0, sc0)

        # we can open non-zero subchannels as soon as we get our first
        # connection, and we can make the Endpoints even earlier
        control_ep = ControlEndpoint(peer_addr0, sc0, self._eventual_queue)
        connect_ep = SubchannelConnectorEndpoint(self, self._host_addr, self._eventual_queue)
        listen_ep = SubchannelListenerEndpoint(self, self._host_addr, self._eventual_queue)
        # TODO: let inbound/outbound create the endpoints, then return them
        # to us
        self._inbound.set_listener_endpoint(listen_ep)

        self._endpoints = EndpointRecord(control_ep, connect_ep, listen_ep)

    def get_endpoints(self):
        return self._endpoints

    def got_dilation_key(self, key):
        assert isinstance(key, bytes)
        self._dilation_key = key

    def got_wormhole_versions(self, their_wormhole_versions):
        # this always happens before received_dilation_message
        self._dilation_version = None
        their_dilation_versions = set(their_wormhole_versions.get("can-dilate", []))
        my_versions = set(DILATION_VERSIONS)
        shared_versions = my_versions.intersection(their_dilation_versions)
        if "1" in shared_versions:
            self._dilation_version = "1"

        # dilation_version is the best mutually-compatible version we have
        # with the peer, or None if we have nothing in common

        if not self._dilation_version:  # "1" or None
            # TODO: be more specific about the error. dilation_version==None
            # means we had no version in common with them, which could either
            # be because they're so old they don't dilate at all, or because
            # they're so new that they no longer accommodate our old version
            self.fail(failure.Failure(OldPeerCannotDilateError()))

        self.start()

    def fail(self, f):
        self._endpoints.control._main_channel_failed(f)
        self._endpoints.connect._main_channel_failed(f)
        self._endpoints.listen._main_channel_failed(f)

    def received_dilation_message(self, plaintext):
        # this receives new in-order DILATE-n payloads, decrypted but not
        # de-JSONed.

        message = bytes_to_dict(plaintext)
        type = message["type"]
        if type == "please":
            self.rx_PLEASE(message)
        elif type == "connection-hints":
            self.rx_HINTS(message)
        elif type == "reconnect":
            self.rx_RECONNECT()
        elif type == "reconnecting":
            self.rx_RECONNECTING()
        else:
            log.err(UnknownDilationMessageType(message))
            return

    def when_stopped(self):
        return self._stopped.when_fired()

    def send_dilation_phase(self, **fields):
        dilation_phase = self._next_dilation_phase
        self._next_dilation_phase += 1
        self._S.send("dilate-%d" % dilation_phase, dict_to_bytes(fields))

    def send_hints(self, hints):  # from Connector
        self.send_dilation_phase(type="connection-hints", hints=hints)

    # forward inbound-ish things to _Inbound

    def subchannel_pauseProducing(self, sc):
        self._inbound.subchannel_pauseProducing(sc)

    def subchannel_resumeProducing(self, sc):
        self._inbound.subchannel_resumeProducing(sc)

    def subchannel_stopProducing(self, sc):
        self._inbound.subchannel_stopProducing(sc)

    def subchannel_local_open(self, scid, sc):
        self._inbound.subchannel_local_open(scid, sc)

    # forward outbound-ish things to _Outbound
    def subchannel_registerProducer(self, sc, producer, streaming):
        self._outbound.subchannel_registerProducer(sc, producer, streaming)

    def subchannel_unregisterProducer(self, sc):
        self._outbound.subchannel_unregisterProducer(sc)

    def send_open(self, scid):
        assert isinstance(scid, int)
        self._queue_and_send(Open, scid)

    def send_data(self, scid, data):
        assert isinstance(scid, int)
        self._queue_and_send(Data, scid, data)

    def send_close(self, scid):
        assert isinstance(scid, int)
        self._queue_and_send(Close, scid)

    def _queue_and_send(self, record_type, *args):
        r = self._outbound.build_record(record_type, *args)
        # Outbound owns the send_record() pipe, so that it can stall new
        # writes after a new connection is made until after all queued
        # messages are written (to preserve ordering).
        self._outbound.queue_and_send_record(r)  # may trigger pauseProducing

    def subchannel_closed(self, scid, sc):
        # let everyone clean up. This happens just after we delivered
        # connectionLost to the Protocol, except for the control channel,
        # which might get connectionLost later after they use ep.connect.
        # TODO: is this inversion a problem?
        self._inbound.subchannel_closed(scid, sc)
        self._outbound.subchannel_closed(scid, sc)

    # our Connector calls these

    def connector_connection_made(self, c):
        self.connection_made()  # state machine update
        self._connection = c
        self._inbound.use_connection(c)
        self._outbound.use_connection(c)  # does c.registerProducer
        if not self._made_first_connection:
            self._made_first_connection = True
            self._endpoints.control._main_channel_ready()
            self._endpoints.connect._main_channel_ready()
            self._endpoints.listen._main_channel_ready()
        pass

    def connector_connection_lost(self):
        self._stop_using_connection()
        if self._my_role is LEADER:
            self.connection_lost_leader()  # state machine
        else:
            self.connection_lost_follower()

    def _stop_using_connection(self):
        # the connection is already lost by this point
        self._connection = None
        self._inbound.stop_using_connection()
        self._outbound.stop_using_connection()  # does c.unregisterProducer

    # from our active Connection

    def got_record(self, r):
        # records with sequence numbers: always ack, ignore old ones
        if isinstance(r, (Open, Data, Close)):
            self.send_ack(r.seqnum)  # always ack, even for old ones
            if self._inbound.is_record_old(r):
                return
            self._inbound.update_ack_watermark(r.seqnum)
            if isinstance(r, Open):
                self._inbound.handle_open(r.scid)
            elif isinstance(r, Data):
                self._inbound.handle_data(r.scid, r.data)
            else:  # isinstance(r, Close)
                self._inbound.handle_close(r.scid)
            return
        if isinstance(r, KCM):
            log.err(UnexpectedKCM())
        elif isinstance(r, Ping):
            self.handle_ping(r.ping_id)
        elif isinstance(r, Pong):
            self.handle_pong(r.ping_id)
        elif isinstance(r, Ack):
            self._outbound.handle_ack(r.resp_seqnum)  # retire queued messages
        else:
            log.err(UnknownMessageType("{}".format(r)))

    # pings, pongs, and acks are not queued
    def send_ping(self, ping_id):
        self._outbound.send_if_connected(Ping(ping_id))

    def send_pong(self, ping_id):
        self._outbound.send_if_connected(Pong(ping_id))

    def send_ack(self, resp_seqnum):
        self._outbound.send_if_connected(Ack(resp_seqnum))

    def handle_ping(self, ping_id):
        self.send_pong(ping_id)

    def handle_pong(self, ping_id):
        # TODO: update is-alive timer
        pass

    # subchannel maintenance
    def allocate_subchannel_id(self):
        scid_num = self._next_subchannel_id
        self._next_subchannel_id += 2
        return scid_num

    # state machine

    @m.state(initial=True)
    def WAITING(self):
        pass  # pragma: no cover

    @m.state()
    def WANTING(self):
        pass  # pragma: no cover

    @m.state()
    def CONNECTING(self):
        pass  # pragma: no cover

    @m.state()
    def CONNECTED(self):
        pass  # pragma: no cover

    @m.state()
    def FLUSHING(self):
        pass  # pragma: no cover

    @m.state()
    def ABANDONING(self):
        pass  # pragma: no cover

    @m.state()
    def LONELY(self):
        pass  # pragma: no cover

    @m.state()
    def STOPPING(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def STOPPED(self):
        pass  # pragma: no cover

    @m.input()
    def start(self):
        pass  # pragma: no cover

    @m.input()
    def rx_PLEASE(self, message):
        pass  # pragma: no cover

    @m.input()  # only sent by Follower
    def rx_HINTS(self, hint_message):
        pass  # pragma: no cover

    @m.input()  # only Leader sends RECONNECT, so only Follower receives it
    def rx_RECONNECT(self):
        pass  # pragma: no cover

    @m.input()  # only Follower sends RECONNECTING, so only Leader receives it
    def rx_RECONNECTING(self):
        pass  # pragma: no cover

    # Connector gives us connection_made()
    @m.input()
    def connection_made(self):
        pass  # pragma: no cover

    # our connection_lost() fires connection_lost_leader or
    # connection_lost_follower depending upon our role. If either side sees a
    # problem with the connection (timeouts, bad authentication) then they
    # just drop it and let connection_lost() handle the cleanup.
    @m.input()
    def connection_lost_leader(self):
        pass  # pragma: no cover

    @m.input()
    def connection_lost_follower(self):
        pass

    @m.input()
    def stop(self):
        pass  # pragma: no cover

    @m.output()
    def send_please(self):
        msg = {
            "type": "please",
            "side": self._my_side,
        }
        if self._dilation_version is not None:
            msg["use-version"] = self._dilation_version
        self.send_dilation_phase(**msg)

    @m.output()
    def choose_role(self, message):
        their_side = message["side"]
        if self._my_side > their_side:
            self._my_role = LEADER
            # scid 0 is reserved for the control channel. the leader uses odd
            # numbers starting with 1
            self._next_subchannel_id = 1
        elif their_side > self._my_side:
            self._my_role = FOLLOWER
            # the follower uses even numbers starting with 2
            self._next_subchannel_id = 2
        else:
            raise ValueError("their side shouldn't be equal: reflection?")

    # these Outputs behave differently for the Leader vs the Follower

    @m.output()
    def start_connecting_ignore_message(self, message):
        del message  # ignored
        return self._start_connecting()

    @m.output()
    def start_connecting(self):
        self._start_connecting()

    def _start_connecting(self):
        assert self._my_role is not None
        assert self._dilation_key is not None
        self._connector = Connector(self._dilation_key,
                                    self._transit_relay_location,
                                    self,
                                    self._reactor, self._eventual_queue,
                                    self._no_listen, self._tor,
                                    self._timing,
                                    self._my_side,  # needed for relay handshake
                                    self._my_role)
        if self._debug_stall_connector:
            # unit tests use this hook to send messages while we know we
            # don't have a connection
            self._eventual_queue.eventually(self._debug_stall_connector, self._connector)
            return
        self._connector.start()

    @m.output()
    def send_reconnect(self):
        self.send_dilation_phase(type="reconnect")  # TODO: generation number?

    @m.output()
    def send_reconnecting(self):
        self.send_dilation_phase(type="reconnecting")  # TODO: generation?

    @m.output()
    def use_hints(self, hint_message):
        hint_objs = filter(lambda h: h,  # ignore None, unrecognizable
                           [parse_hint(hs) for hs in hint_message["hints"]])
        hint_objs = list(hint_objs)
        self._connector.got_hints(hint_objs)

    @m.output()
    def stop_connecting(self):
        self._connector.stop()

    @m.output()
    def abandon_connection(self):
        # we think we're still connected, but the Leader disagrees. Or we've
        # been told to shut down.
        self._connection.disconnect()  # let connection_lost do cleanup

    @m.output()
    def notify_stopped(self):
        self._stopped.fire(None)

    # We are born WAITING after the local app calls w.dilate(). We enter
    # WANTING (and send a PLEASE) when we learn of a mutually-compatible
    # dilation_version.
    WAITING.upon(start, enter=WANTING, outputs=[send_please])

    # we start CONNECTING when we get rx_PLEASE
    WANTING.upon(rx_PLEASE, enter=CONNECTING,
                 outputs=[choose_role, start_connecting_ignore_message])

    CONNECTING.upon(connection_made, enter=CONNECTED, outputs=[])

    # Leader
    CONNECTED.upon(connection_lost_leader, enter=FLUSHING,
                   outputs=[send_reconnect])
    FLUSHING.upon(rx_RECONNECTING, enter=CONNECTING,
                  outputs=[start_connecting])

    # Follower
    # if we notice a lost connection, just wait for the Leader to notice too
    CONNECTED.upon(connection_lost_follower, enter=LONELY, outputs=[])
    LONELY.upon(rx_RECONNECT, enter=CONNECTING,
                outputs=[send_reconnecting, start_connecting])
    # but if they notice it first, abandon our (seemingly functional)
    # connection, then tell them that we're ready to try again
    CONNECTED.upon(rx_RECONNECT, enter=ABANDONING, outputs=[abandon_connection])
    ABANDONING.upon(connection_lost_follower, enter=CONNECTING,
                    outputs=[send_reconnecting, start_connecting])
    # and if they notice a problem while we're still connecting, abandon our
    # incomplete attempt and try again. in this case we don't have to wait
    # for a connection to finish shutdown
    CONNECTING.upon(rx_RECONNECT, enter=CONNECTING,
                    outputs=[stop_connecting,
                             send_reconnecting,
                             start_connecting])

    # rx_HINTS never changes state, they're just accepted or ignored
    WANTING.upon(rx_HINTS, enter=WANTING, outputs=[])  # too early
    CONNECTING.upon(rx_HINTS, enter=CONNECTING, outputs=[use_hints])
    CONNECTED.upon(rx_HINTS, enter=CONNECTED, outputs=[])  # too late, ignore
    FLUSHING.upon(rx_HINTS, enter=FLUSHING, outputs=[])  # stale, ignore
    LONELY.upon(rx_HINTS, enter=LONELY, outputs=[])  # stale, ignore
    ABANDONING.upon(rx_HINTS, enter=ABANDONING, outputs=[])  # shouldn't happen
    STOPPING.upon(rx_HINTS, enter=STOPPING, outputs=[])

    WAITING.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    WANTING.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    CONNECTING.upon(stop, enter=STOPPED, outputs=[stop_connecting, notify_stopped])
    CONNECTED.upon(stop, enter=STOPPING, outputs=[abandon_connection])
    ABANDONING.upon(stop, enter=STOPPING, outputs=[])
    FLUSHING.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    LONELY.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    STOPPING.upon(connection_lost_leader, enter=STOPPED, outputs=[notify_stopped])
    STOPPING.upon(connection_lost_follower, enter=STOPPED, outputs=[notify_stopped])


@attrs
@implementer(IDilator)
class Dilator(object):
    """I launch the dilation process.

    I am created with every Wormhole (regardless of whether .dilate()
    was called or not), and I handle the initial phase of dilation,
    before we know whether we'll be the Leader or the Follower. Once we
    hear the other side's VERSION message (which tells us that we have a
    connection, they are capable of dilating, and which side we're on),
    then we build a Manager and hand control to it.
    """

    _reactor = attrib()
    _eventual_queue = attrib()
    _cooperator = attrib()

    def __attrs_post_init__(self):
        self._manager = None
        self._pending_dilation_key = None
        self._pending_wormhole_versions = None
        self._pending_inbound_dilate_messages = deque()

    def wire(self, sender, terminator):
        self._S = ISend(sender)
        self._T = ITerminator(terminator)

    # this is the primary entry point, called when w.dilate() is invoked
    def dilate(self, transit_relay_location=None, no_listen=False):
        if not self._manager:
            # build the manager right away, and tell it later when the
            # VERSIONS message arrives, and also when the dilation_key is set
            my_dilation_side = make_side()
            m = Manager(self._S, my_dilation_side,
                        transit_relay_location,
                        self._reactor, self._eventual_queue,
                        self._cooperator, no_listen)
            self._manager = m
            if self._pending_dilation_key is not None:
                m.got_dilation_key(self._pending_dilation_key)
            if self._pending_wormhole_versions:
                m.got_wormhole_versions(self._pending_wormhole_versions)
            while self._pending_inbound_dilate_messages:
                plaintext = self._pending_inbound_dilate_messages.popleft()
                m.received_dilation_message(plaintext)
        return self._manager.get_endpoints()

    # Called by Terminator after everything else (mailbox, nameplate, server
    # connection) has shut down. Expects to fire T.stoppedD() when Dilator is
    # stopped too.
    def stop(self):
        if self._manager:
            self._manager.stop()
            # TODO: avoid Deferreds for control flow, hard to serialize
            self._manager.when_stopped().addCallback(lambda _: self._T.stoppedD())
        else:
            self._T.stoppedD()
            return
        # TODO: tolerate multiple calls

    # from Boss

    def got_key(self, key):
        # TODO: verify this happens before got_wormhole_versions, or add a gate
        # to tolerate either ordering
        purpose = b"dilation-v1"
        LENGTH = 32  # TODO: whatever Noise wants, I guess
        dilation_key = derive_key(key, purpose, LENGTH)
        if self._manager:
            self._manager.got_dilation_key(dilation_key)
        else:
            self._pending_dilation_key = dilation_key

    def got_wormhole_versions(self, their_wormhole_versions):
        if self._manager:
            self._manager.got_wormhole_versions(their_wormhole_versions)
        else:
            self._pending_wormhole_versions = their_wormhole_versions

    def received_dilate(self, plaintext):
        if not self._manager:
            self._pending_inbound_dilate_messages.append(plaintext)
        else:
            self._manager.received_dilation_message(plaintext)



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.task import Clock
from twisted.python.failure import Failure
from ..._interfaces import ISubChannel
from ...eventual import EventualQueue
from ..._dilation.subchannel import (ControlEndpoint,
                                     SubchannelConnectorEndpoint,
                                     SubchannelListenerEndpoint,
                                     SubchannelListeningPort,
                                     _WormholeAddress, _SubchannelAddress,
                                     SingleUseEndpointError)
from .common import mock_manager


class CannotDilateError(Exception):
    pass


class Control(unittest.TestCase):
    def test_early_succeed(self):
        # ep.connect() is called before dilation can proceed
        scid0 = 0
        peeraddr = _SubchannelAddress(scid0)
        sc0 = mock.Mock()
        alsoProvides(sc0, ISubChannel)
        eq = EventualQueue(Clock())
        ep = ControlEndpoint(peeraddr, sc0, eq)

        f = mock.Mock()
        p = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        d = ep.connect(f)
        self.assertNoResult(d)

        ep._main_channel_ready()
        eq.flush_sync()

        self.assertIdentical(self.successResultOf(d), p)
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr)])
        self.assertEqual(sc0.mock_calls, [mock.call._set_protocol(p),
                                          mock.call._deliver_queued_data()])
        self.assertEqual(p.mock_calls, [mock.call.makeConnection(sc0)])

        d = ep.connect(f)
        self.failureResultOf(d, SingleUseEndpointError)

    def test_early_fail(self):
        # ep.connect() is called before dilation is abandoned
        scid0 = 0
        peeraddr = _SubchannelAddress(scid0)
        sc0 = mock.Mock()
        alsoProvides(sc0, ISubChannel)
        eq = EventualQueue(Clock())
        ep = ControlEndpoint(peeraddr, sc0, eq)

        f = mock.Mock()
        p = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        d = ep.connect(f)
        self.assertNoResult(d)

        ep._main_channel_failed(Failure(CannotDilateError()))
        eq.flush_sync()

        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])
        self.assertEqual(sc0.mock_calls, [])

        d = ep.connect(f)
        self.failureResultOf(d, SingleUseEndpointError)

    def test_late_succeed(self):
        # dilation can proceed, then ep.connect() is called
        scid0 = 0
        peeraddr = _SubchannelAddress(scid0)
        sc0 = mock.Mock()
        alsoProvides(sc0, ISubChannel)
        eq = EventualQueue(Clock())
        ep = ControlEndpoint(peeraddr, sc0, eq)

        ep._main_channel_ready()

        f = mock.Mock()
        p = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        d = ep.connect(f)
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d), p)
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr)])
        self.assertEqual(sc0.mock_calls, [mock.call._set_protocol(p),
                                          mock.call._deliver_queued_data()])
        self.assertEqual(p.mock_calls, [mock.call.makeConnection(sc0)])

        d = ep.connect(f)
        self.failureResultOf(d, SingleUseEndpointError)

    def test_late_fail(self):
        # dilation is abandoned, then ep.connect() is called
        scid0 = 0
        peeraddr = _SubchannelAddress(scid0)
        sc0 = mock.Mock()
        alsoProvides(sc0, ISubChannel)
        eq = EventualQueue(Clock())
        ep = ControlEndpoint(peeraddr, sc0, eq)

        ep._main_channel_failed(Failure(CannotDilateError()))

        f = mock.Mock()
        p = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        d = ep.connect(f)
        eq.flush_sync()

        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])
        self.assertEqual(sc0.mock_calls, [])

        d = ep.connect(f)
        self.failureResultOf(d, SingleUseEndpointError)


class Endpoints(unittest.TestCase):
    def OFFassert_makeConnection(self, mock_calls):
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][0], "makeConnection")
        self.assertEqual(len(mock_calls[0][1]), 1)
        return mock_calls[0][1][0]


class Connector(unittest.TestCase):
    def test_early_succeed(self):
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        peeraddr = _SubchannelAddress(0)
        eq = EventualQueue(Clock())
        ep = SubchannelConnectorEndpoint(m, hostaddr, eq)

        f = mock.Mock()
        p = mock.Mock()
        t = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        with mock.patch("wormhole._dilation.subchannel.SubChannel",
                        return_value=t) as sc:
            d = ep.connect(f)
            eq.flush_sync()
            self.assertNoResult(d)
            ep._main_channel_ready()
            eq.flush_sync()

        self.assertIdentical(self.successResultOf(d), p)
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr)])
        self.assertEqual(sc.mock_calls, [mock.call(0, m, hostaddr, peeraddr)])
        self.assertEqual(t.mock_calls, [mock.call._set_protocol(p)])
        self.assertEqual(p.mock_calls, [mock.call.makeConnection(t)])

    def test_early_fail(self):
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelConnectorEndpoint(m, hostaddr, eq)

        f = mock.Mock()
        p = mock.Mock()
        t = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        with mock.patch("wormhole._dilation.subchannel.SubChannel",
                        return_value=t) as sc:
            d = ep.connect(f)
            eq.flush_sync()
            self.assertNoResult(d)
            ep._main_channel_failed(Failure(CannotDilateError()))
            eq.flush_sync()

        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])
        self.assertEqual(sc.mock_calls, [])
        self.assertEqual(t.mock_calls, [])

    def test_late_succeed(self):
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        peeraddr = _SubchannelAddress(0)
        eq = EventualQueue(Clock())
        ep = SubchannelConnectorEndpoint(m, hostaddr, eq)
        ep._main_channel_ready()

        f = mock.Mock()
        p = mock.Mock()
        t = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        with mock.patch("wormhole._dilation.subchannel.SubChannel",
                        return_value=t) as sc:
            d = ep.connect(f)
            eq.flush_sync()

        self.assertIdentical(self.successResultOf(d), p)
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr)])
        self.assertEqual(sc.mock_calls, [mock.call(0, m, hostaddr, peeraddr)])
        self.assertEqual(t.mock_calls, [mock.call._set_protocol(p)])
        self.assertEqual(p.mock_calls, [mock.call.makeConnection(t)])

    def test_late_fail(self):
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelConnectorEndpoint(m, hostaddr, eq)
        ep._main_channel_failed(Failure(CannotDilateError()))

        f = mock.Mock()
        p = mock.Mock()
        t = mock.Mock()
        f.buildProtocol = mock.Mock(return_value=p)
        with mock.patch("wormhole._dilation.subchannel.SubChannel",
                        return_value=t) as sc:
            d = ep.connect(f)
            eq.flush_sync()

        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])
        self.assertEqual(sc.mock_calls, [])
        self.assertEqual(t.mock_calls, [])


class Listener(unittest.TestCase):
    def test_early_succeed(self):
        # listen, main_channel_ready, got_open, got_open
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelListenerEndpoint(m, hostaddr, eq)

        f = mock.Mock()
        p1 = mock.Mock()
        p2 = mock.Mock()
        f.buildProtocol = mock.Mock(side_effect=[p1, p2])

        d = ep.listen(f)
        eq.flush_sync()
        self.assertNoResult(d)
        self.assertEqual(f.buildProtocol.mock_calls, [])

        ep._main_channel_ready()
        eq.flush_sync()
        lp = self.successResultOf(d)
        self.assertIsInstance(lp, SubchannelListeningPort)

        self.assertEqual(lp.getHost(), hostaddr)
        # TODO: IListeningPort says we must provide this, but I don't know
        # that anyone would ever call it.
        lp.startListening()

        t1 = mock.Mock()
        peeraddr1 = _SubchannelAddress(1)
        ep._got_open(t1, peeraddr1)

        self.assertEqual(t1.mock_calls, [mock.call._set_protocol(p1),
                                         mock.call._deliver_queued_data()])
        self.assertEqual(p1.mock_calls, [mock.call.makeConnection(t1)])
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr1)])

        t2 = mock.Mock()
        peeraddr2 = _SubchannelAddress(2)
        ep._got_open(t2, peeraddr2)

        self.assertEqual(t2.mock_calls, [mock.call._set_protocol(p2),
                                         mock.call._deliver_queued_data()])
        self.assertEqual(p2.mock_calls, [mock.call.makeConnection(t2)])
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr1),
                                                      mock.call(peeraddr2)])

        lp.stopListening()  # TODO: should this do more?

    def test_early_fail(self):
        # listen, main_channel_fail
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelListenerEndpoint(m, hostaddr, eq)

        f = mock.Mock()
        p1 = mock.Mock()
        p2 = mock.Mock()
        f.buildProtocol = mock.Mock(side_effect=[p1, p2])

        d = ep.listen(f)
        eq.flush_sync()
        self.assertNoResult(d)

        ep._main_channel_failed(Failure(CannotDilateError()))
        eq.flush_sync()
        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])

    def test_late_succeed(self):
        # main_channel_ready, got_open, listen, got_open
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelListenerEndpoint(m, hostaddr, eq)
        ep._main_channel_ready()

        f = mock.Mock()
        p1 = mock.Mock()
        p2 = mock.Mock()
        f.buildProtocol = mock.Mock(side_effect=[p1, p2])

        t1 = mock.Mock()
        peeraddr1 = _SubchannelAddress(1)
        ep._got_open(t1, peeraddr1)
        eq.flush_sync()

        self.assertEqual(t1.mock_calls, [])
        self.assertEqual(p1.mock_calls, [])

        d = ep.listen(f)
        eq.flush_sync()
        lp = self.successResultOf(d)
        self.assertIsInstance(lp, SubchannelListeningPort)
        self.assertEqual(lp.getHost(), hostaddr)
        lp.startListening()

        # TODO: assert makeConnection is called *before* _deliver_queued_data
        self.assertEqual(t1.mock_calls, [mock.call._set_protocol(p1),
                                         mock.call._deliver_queued_data()])
        self.assertEqual(p1.mock_calls, [mock.call.makeConnection(t1)])
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr1)])

        t2 = mock.Mock()
        peeraddr2 = _SubchannelAddress(2)
        ep._got_open(t2, peeraddr2)

        self.assertEqual(t2.mock_calls, [mock.call._set_protocol(p2),
                                         mock.call._deliver_queued_data()])
        self.assertEqual(p2.mock_calls, [mock.call.makeConnection(t2)])
        self.assertEqual(f.buildProtocol.mock_calls, [mock.call(peeraddr1),
                                                      mock.call(peeraddr2)])

        lp.stopListening()  # TODO: should this do more?

    def test_late_fail(self):
        # main_channel_fail, listen
        m = mock_manager()
        m.allocate_subchannel_id = mock.Mock(return_value=0)
        hostaddr = _WormholeAddress()
        eq = EventualQueue(Clock())
        ep = SubchannelListenerEndpoint(m, hostaddr, eq)
        ep._main_channel_failed(Failure(CannotDilateError()))

        f = mock.Mock()
        p1 = mock.Mock()
        p2 = mock.Mock()
        f.buildProtocol = mock.Mock(side_effect=[p1, p2])

        d = ep.listen(f)
        eq.flush_sync()
        self.failureResultOf(d).check(CannotDilateError)
        self.assertEqual(f.buildProtocol.mock_calls, [])



import builtins
import io
import os
import re
import stat
import sys
import zipfile
from textwrap import dedent, fill

from click import UsageError
from click.testing import CliRunner
from humanize import naturalsize
from twisted.internet import endpoints, reactor
from twisted.internet.defer import gatherResults, inlineCallbacks, returnValue, CancelledError
from twisted.internet.error import ConnectionRefusedError
from twisted.internet.utils import getProcessOutputAndValue
from twisted.python import log, procutils
from twisted.trial import unittest
from zope.interface import implementer

from unittest import mock

from .. import __version__
from .._interfaces import ITorManager
from ..cli import cli, cmd_receive, cmd_send, welcome
from ..errors import (ServerConnectionError, TransferError,
                      UnsendableFileError, WelcomeError, WrongPasswordError)
from .common import ServerBase, config


def build_offer(args):
    s = cmd_send.Sender(args, None)
    return s._build_offer()


class OfferData(unittest.TestCase):
    def setUp(self):
        self._things_to_delete = []
        self.cfg = cfg = config("send")
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

    def tearDown(self):
        for fn in self._things_to_delete:
            if os.path.exists(fn):
                os.unlink(fn)
        del self.cfg

    def test_text(self):
        self.cfg.text = message = "blah blah blah ponies"
        d, fd_to_send = build_offer(self.cfg)

        self.assertIn("message", d)
        self.assertNotIn("file", d)
        self.assertNotIn("directory", d)
        self.assertEqual(d["message"], message)
        self.assertEqual(fd_to_send, None)

    def test_file(self):
        self.cfg.what = filename = "my file"
        message = b"yay ponies\n"
        send_dir = self.mktemp()
        os.mkdir(send_dir)
        abs_filename = os.path.join(send_dir, filename)
        with open(abs_filename, "wb") as f:
            f.write(message)

        self.cfg.cwd = send_dir
        d, fd_to_send = build_offer(self.cfg)

        self.assertNotIn("message", d)
        self.assertIn("file", d)
        self.assertNotIn("directory", d)
        self.assertEqual(d["file"]["filesize"], len(message))
        self.assertEqual(d["file"]["filename"], filename)
        self.assertEqual(fd_to_send.tell(), 0)
        self.assertEqual(fd_to_send.read(), message)

    def _create_broken_symlink(self):
        if not hasattr(os, 'symlink'):
            raise unittest.SkipTest("host OS does not support symlinks")

        parent_dir = self.mktemp()
        os.mkdir(parent_dir)
        send_dir = "dirname"
        os.mkdir(os.path.join(parent_dir, send_dir))
        os.symlink('/non/existent/file',
                   os.path.join(parent_dir, send_dir, 'linky'))

        send_dir_arg = send_dir
        self.cfg.what = send_dir_arg
        self.cfg.cwd = parent_dir

    def test_broken_symlink_raises_err(self):
        self._create_broken_symlink()
        self.cfg.ignore_unsendable_files = False
        e = self.assertRaises(UnsendableFileError, build_offer, self.cfg)

        # On english distributions of Linux, this will be
        # "linky: No such file or directory", but the error may be
        # different on Windows and other locales and/or Unix variants, so
        # we'll just assert the part we know about.
        self.assertIn("linky: ", str(e))

    def test_broken_symlink_is_ignored(self):
        self._create_broken_symlink()
        self.cfg.ignore_unsendable_files = True
        d, fd_to_send = build_offer(self.cfg)
        self.assertIn('(ignoring error)', self.cfg.stderr.getvalue())
        self.assertEqual(d['directory']['numfiles'], 0)
        self.assertEqual(d['directory']['numbytes'], 0)

    def test_missing_file(self):
        self.cfg.what = filename = "missing"
        send_dir = self.mktemp()
        os.mkdir(send_dir)
        self.cfg.cwd = send_dir

        e = self.assertRaises(TransferError, build_offer, self.cfg)
        self.assertEqual(
            str(e), "Cannot send: no file/directory named '%s'" % filename)

    def _do_test_directory(self, addslash):
        parent_dir = self.mktemp()
        os.mkdir(parent_dir)
        send_dir = "dirname"
        os.mkdir(os.path.join(parent_dir, send_dir))
        ponies = [str(i) for i in range(5)]
        for p in ponies:
            with open(os.path.join(parent_dir, send_dir, p), "wb") as f:
                f.write(("%s ponies\n" % p).encode("ascii"))

        send_dir_arg = send_dir
        if addslash:
            send_dir_arg += os.sep
        self.cfg.what = send_dir_arg
        self.cfg.cwd = parent_dir

        d, fd_to_send = build_offer(self.cfg)

        self.assertNotIn("message", d)
        self.assertNotIn("file", d)
        self.assertIn("directory", d)
        self.assertEqual(d["directory"]["dirname"], send_dir)
        self.assertEqual(d["directory"]["mode"], "zipfile/deflated")
        self.assertEqual(d["directory"]["numfiles"], 5)
        self.assertIn("numbytes", d["directory"])
        self.assertIsInstance(d["directory"]["numbytes"], int)

        zdata = b"".join(fd_to_send)
        self.assertEqual(len(zdata), d["directory"]["zipsize"])
        with zipfile.ZipFile(io.BytesIO(zdata), "r") as zf:
            zipnames = zf.namelist()
            self.assertEqual(list(sorted(ponies)), list(sorted(zipnames)))
            for name in zipnames:
                contents = zf.open(name, "r").read()
                self.assertEqual(("%s ponies\n" % name).encode("ascii"),
                                 contents)

    def test_directory(self):
        return self._do_test_directory(addslash=False)

    def test_directory_addslash(self):
        return self._do_test_directory(addslash=True)

    def test_unknown(self):
        self.cfg.what = filename = "unknown"
        send_dir = self.mktemp()
        os.mkdir(send_dir)
        abs_filename = os.path.abspath(os.path.join(send_dir, filename))
        self.cfg.cwd = send_dir

        try:
            os.mkfifo(abs_filename)
        except AttributeError:
            raise unittest.SkipTest("is mkfifo supported on this platform?")

        # Delete the named pipe for the sake of users who might run "pip
        # wheel ." in this directory later. That command wants to copy
        # everything into a tempdir before building a wheel, and the
        # shutil.copy_tree() is uses can't handle the named pipe.
        self._things_to_delete.append(abs_filename)

        self.assertFalse(os.path.isfile(abs_filename))
        self.assertFalse(os.path.isdir(abs_filename))

        e = self.assertRaises(TypeError, build_offer, self.cfg)
        self.assertEqual(
            str(e), "'%s' is neither file nor directory" % filename)

    def test_symlink(self):
        if not hasattr(os, 'symlink'):
            raise unittest.SkipTest("host OS does not support symlinks")
        # build A/B1 -> B2 (==A/B2), and A/B2/C.txt
        parent_dir = self.mktemp()
        os.mkdir(parent_dir)
        os.mkdir(os.path.join(parent_dir, "B2"))
        with open(os.path.join(parent_dir, "B2", "C.txt"), "wb") as f:
            f.write(b"success")
        os.symlink("B2", os.path.join(parent_dir, "B1"))
        # now send "B1/C.txt" from A, and it should get the right file
        self.cfg.cwd = parent_dir
        self.cfg.what = os.path.join("B1", "C.txt")
        d, fd_to_send = build_offer(self.cfg)
        self.assertEqual(d["file"]["filename"], "C.txt")
        self.assertEqual(fd_to_send.read(), b"success")

    def test_symlink_collapse(self):
        if not hasattr(os, 'symlink'):
            raise unittest.SkipTest("host OS does not support symlinks")
        # build A/B1, A/B1/D.txt
        # A/B2/C2, A/B2/D.txt
        # symlink A/B1/C1 -> A/B2/C2
        parent_dir = self.mktemp()
        os.mkdir(parent_dir)
        os.mkdir(os.path.join(parent_dir, "B1"))
        with open(os.path.join(parent_dir, "B1", "D.txt"), "wb") as f:
            f.write(b"fail")
        os.mkdir(os.path.join(parent_dir, "B2"))
        os.mkdir(os.path.join(parent_dir, "B2", "C2"))
        with open(os.path.join(parent_dir, "B2", "D.txt"), "wb") as f:
            f.write(b"success")
        os.symlink(
            os.path.abspath(os.path.join(parent_dir, "B2", "C2")),
            os.path.join(parent_dir, "B1", "C1"))
        # Now send "B1/C1/../D.txt" from A. The correct traversal will be:
        # * start: A
        # * B1: A/B1
        # * C1: follow symlink to A/B2/C2
        # * ..: climb to A/B2
        # * D.txt: open A/B2/D.txt, which contains "success"
        # If the code mistakenly uses normpath(), it would do:
        # * normpath turns B1/C1/../D.txt into B1/D.txt
        # * start: A
        # * B1: A/B1
        # * D.txt: open A/B1/D.txt , which contains "fail"
        self.cfg.cwd = parent_dir
        self.cfg.what = os.path.join("B1", "C1", os.pardir, "D.txt")
        d, fd_to_send = build_offer(self.cfg)
        self.assertEqual(d["file"]["filename"], "D.txt")
        self.assertEqual(fd_to_send.read(), b"success")

    if os.name == "nt":
        test_symlink_collapse.todo = "host OS has broken os.path.realpath()"
        # ntpath.py's realpath() is built out of normpath(), and does not
        # follow symlinks properly, so this test always fails. "wormhole send
        # PATH" on windows will do the wrong thing. See
        # https://bugs.python.org/issue9949" for details. I'm making this a
        # TODO instead of a SKIP because 1: this causes an observable
        # misbehavior (albeit in rare circumstances), 2: it probably used to
        # work (sometimes, but not in #251). See cmd_send.py for more notes.


class LocaleFinder:
    def __init__(self):
        self._run_once = False

    @inlineCallbacks
    def find_utf8_locale(self):
        if sys.platform == "win32":
            returnValue("en_US.UTF-8")
        if self._run_once:
            returnValue(self._best_locale)
        self._best_locale = yield self._find_utf8_locale()
        self._run_once = True
        returnValue(self._best_locale)

    @inlineCallbacks
    def _find_utf8_locale(self):
        # Click really wants to be running under a unicode-capable locale,
        # especially on python3. macOS has en-US.UTF-8 but not C.UTF-8, and
        # most linux boxes have C.UTF-8 but not en-US.UTF-8 . For tests,
        # figure out which one is present and use that. For runtime, it's a
        # mess, as really the user must take responsibility for setting their
        # locale properly. I'm thinking of abandoning Click and going back to
        # twisted.python.usage to avoid this problem in the future.
        (out, err, rc) = yield getProcessOutputAndValue("locale", ["-a"])
        if rc != 0:
            log.msg("error running 'locale -a', rc=%s" % (rc, ))
            log.msg("stderr: %s" % (err, ))
            returnValue(None)
        out = out.decode("utf-8")  # make sure we get a string
        utf8_locales = {}
        for locale in out.splitlines():
            locale = locale.strip()
            if locale.lower().endswith((".utf-8", ".utf8")):
                utf8_locales[locale.lower()] = locale
        for wanted in ["C.utf8", "C.UTF-8", "en_US.utf8", "en_US.UTF-8"]:
            if wanted.lower() in utf8_locales:
                returnValue(utf8_locales[wanted.lower()])
        if utf8_locales:
            returnValue(list(utf8_locales.values())[0])
        returnValue(None)


locale_finder = LocaleFinder()


class ScriptsBase:
    def find_executable(self):
        # to make sure we're running the right executable (in a virtualenv),
        # we require that our "wormhole" lives in the same directory as our
        # "python"
        locations = procutils.which("wormhole")
        if not locations:
            raise unittest.SkipTest("unable to find 'wormhole' in $PATH")
        wormhole = locations[0]
        if (os.path.dirname(os.path.abspath(wormhole)) != os.path.dirname(
                sys.executable)):
            log.msg("locations: %s" % (locations, ))
            log.msg("sys.executable: %s" % (sys.executable, ))
            raise unittest.SkipTest(
                "found the wrong 'wormhole' in $PATH: %s %s" %
                (wormhole, sys.executable))
        return wormhole

    @inlineCallbacks
    def is_runnable(self):
        # One property of Versioneer is that many changes to the source tree
        # (making a commit, dirtying a previously-clean tree) will change the
        # version string. Entrypoint scripts frequently insist upon importing
        # a library version that matches the script version (whatever was
        # reported when 'pip install' was run), and throw a
        # DistributionNotFound error when they don't match. This is really
        # annoying in a workspace created with "pip install -e .", as you
        # must re-run pip after each commit.

        # So let's report just one error in this case (from test_version),
        # and skip the other tests that we know will fail.

        # Setting LANG/LC_ALL to a unicode-capable locale is necessary to
        # convince Click to not complain about a forced-ascii locale. My
        # apologies to folks who want to run tests on a machine that doesn't
        # have the C.UTF-8 locale installed.
        locale = yield locale_finder.find_utf8_locale()
        if not locale:
            raise unittest.SkipTest("unable to find UTF-8 locale")
        locale_env = dict(LC_ALL=locale, LANG=locale)
        wormhole = self.find_executable()
        res = yield getProcessOutputAndValue(
            wormhole, ["--version"], env=locale_env)
        out, err, rc = res
        if rc != 0:
            log.msg("wormhole not runnable in this tree:")
            log.msg("out", out)
            log.msg("err", err)
            log.msg("rc", rc)
            raise unittest.SkipTest("wormhole is not runnable in this tree")
        returnValue(locale_env)


class ScriptVersion(ServerBase, ScriptsBase, unittest.TestCase):
    # we need Twisted to run the server, but we run the sender and receiver
    # with deferToThread()

    @inlineCallbacks
    def test_version(self):
        # "wormhole" must be on the path, so e.g. "pip install -e ." in a
        # virtualenv. This guards against an environment where the tests
        # below might run the wrong executable.
        self.maxDiff = None
        wormhole = self.find_executable()
        # we must pass on the environment so that "something" doesn't
        # get sad about UTF8 vs. ascii encodings
        out, err, rc = yield getProcessOutputAndValue(
            wormhole, ["--version"], env=os.environ)
        err = err.decode("utf-8")
        if "DistributionNotFound" in err:
            log.msg("stderr was %s" % err)
            last = err.strip().split("\n")[-1]
            self.fail("wormhole not runnable: %s" % last)
        ver = out.decode("utf-8") or err
        self.failUnlessEqual(ver.strip(),
                             "magic-wormhole {}".format(__version__))
        self.failUnlessEqual(rc, 0)


@implementer(ITorManager)
class FakeTor:
    # use normal endpoints, but record the fact that we were asked
    def __init__(self):
        self.endpoints = []

    def stream_via(self, host, port, tls=False):
        self.endpoints.append((host, port, tls))
        return endpoints.HostnameEndpoint(reactor, host, port)


def strip_deprecations(stderr, NL):
    lines = [line
             for line in stderr.split(NL)
             if not ("Python 2 is no longer supported" in line or
                     "from cryptography import utils" in line or
                     "support will be dropped in the next release of cryptography" in line
                     )
             ]
    return NL.join(lines)


class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
    # we need Twisted to run the server, but we run the sender and receiver
    # with deferToThread()

    @inlineCallbacks
    def setUp(self):
        self._env = yield self.is_runnable()
        yield ServerBase.setUp(self)

    @inlineCallbacks
    def _do_test(self,
                 as_subprocess=False,
                 mode="text",
                 addslash=False,
                 override_filename=False,
                 fake_tor=False,
                 overwrite=False,
                 mock_accept=False,
                 verify=False):
        assert mode in ("text", "file", "empty-file", "directory", "slow-text",
                        "slow-sender-text")
        if fake_tor:
            assert not as_subprocess
        send_cfg = config("send")
        recv_cfg = config("receive")
        message = "blah blah blah ponies"

        for cfg in [send_cfg, recv_cfg]:
            cfg.hide_progress = True
            cfg.relay_url = self.relayurl
            cfg.transit_helper = ""
            cfg.listen = True
            cfg.code = u"1-abc"
            cfg.stdout = io.StringIO()
            cfg.stderr = io.StringIO()
            cfg.verify = verify

        send_dir = self.mktemp()
        os.mkdir(send_dir)
        receive_dir = self.mktemp()
        os.mkdir(receive_dir)

        if mode in ("text", "slow-text", "slow-sender-text"):
            send_cfg.text = message

        elif mode in ("file", "empty-file"):
            if mode == "empty-file":
                message = ""
            send_filename = u"testfil\u00EB"  # e-with-diaeresis
            with open(os.path.join(send_dir, send_filename), "w") as f:
                f.write(message)
            send_cfg.what = send_filename
            receive_filename = send_filename

            recv_cfg.accept_file = False if mock_accept else True
            if override_filename:
                recv_cfg.output_file = receive_filename = u"outfile"
            if overwrite:
                recv_cfg.output_file = receive_filename
                existing_file = os.path.join(receive_dir, receive_filename)
                with open(existing_file, 'w') as f:
                    f.write('pls overwrite me')

        elif mode == "directory":
            # $send_dir/
            # $send_dir/middle/
            # $send_dir/middle/$dirname/
            # $send_dir/middle/$dirname/[12345]
            # cd $send_dir && wormhole send middle/$dirname
            # cd $receive_dir && wormhole receive
            # expect: $receive_dir/$dirname/[12345]

            send_dirname = u"testdir"

            def message(i):
                return "test message %d\n" % i

            os.mkdir(os.path.join(send_dir, u"middle"))
            source_dir = os.path.join(send_dir, u"middle", send_dirname)
            os.mkdir(source_dir)
            modes = {}
            for i in range(5):
                path = os.path.join(source_dir, str(i))
                with open(path, "w") as f:
                    f.write(message(i))
                if i == 3:
                    os.chmod(path, 0o755)
                modes[i] = stat.S_IMODE(os.stat(path).st_mode)
            send_dirname_arg = os.path.join(u"middle", send_dirname)
            if addslash:
                send_dirname_arg += os.sep
            send_cfg.what = send_dirname_arg
            receive_dirname = send_dirname

            recv_cfg.accept_file = False if mock_accept else True
            if override_filename:
                recv_cfg.output_file = receive_dirname = u"outdir"
            if overwrite:
                recv_cfg.output_file = receive_dirname
                os.mkdir(os.path.join(receive_dir, receive_dirname))

        if as_subprocess:
            wormhole_bin = self.find_executable()
            if send_cfg.text:
                content_args = ['--text', send_cfg.text]
            elif send_cfg.what:
                content_args = [send_cfg.what]

            # raise the rx KEY_TIMER to some large number here, to avoid
            # spurious test failures on hosts that are slow enough to trigger
            # the "Waiting for sender..." pacifier message. We can do in
            # not-as_subprocess, because we can directly patch the value before
            # running the receiver. But we can't patch across the subprocess
            # boundary, so we use an environment variable.
            env = self._env.copy()
            env["_MAGIC_WORMHOLE_TEST_KEY_TIMER"] = "999999"
            env["_MAGIC_WORMHOLE_TEST_VERIFY_TIMER"] = "999999"
            send_args = [
                '--relay-url',
                self.relayurl,
                '--transit-helper',
                '',
                'send',
                '--hide-progress',
                '--code',
                send_cfg.code,
            ] + content_args

            send_d = getProcessOutputAndValue(
                wormhole_bin,
                send_args,
                path=send_dir,
                env=env,
            )
            recv_args = [
                '--relay-url',
                self.relayurl,
                '--transit-helper',
                '',
                'receive',
                '--hide-progress',
                '--accept-file',
                recv_cfg.code,
            ]
            if override_filename:
                recv_args.extend(['-o', receive_filename])

            receive_d = getProcessOutputAndValue(
                wormhole_bin,
                recv_args,
                path=receive_dir,
                env=env,
            )

            (send_res, receive_res) = yield gatherResults([send_d, receive_d],
                                                          True)
            send_stdout = send_res[0].decode("utf-8")
            send_stderr = send_res[1].decode("utf-8")
            send_rc = send_res[2]
            receive_stdout = receive_res[0].decode("utf-8")
            receive_stderr = receive_res[1].decode("utf-8")
            receive_rc = receive_res[2]
            NL = os.linesep
            send_stderr = strip_deprecations(send_stderr, NL)
            receive_stderr = strip_deprecations(receive_stderr, NL)
            self.assertEqual((send_rc, receive_rc), (0, 0),
                             (send_res, receive_res))
        else:
            send_cfg.cwd = send_dir
            recv_cfg.cwd = receive_dir

            if fake_tor:
                send_cfg.tor = True
                send_cfg.transit_helper = self.transit
                tx_tm = FakeTor()
                with mock.patch(
                        "wormhole.tor_manager.get_tor",
                        return_value=tx_tm,
                ) as mtx_tm:
                    send_d = cmd_send.send(send_cfg)

                recv_cfg.tor = True
                recv_cfg.transit_helper = self.transit
                rx_tm = FakeTor()
                with mock.patch(
                        "wormhole.tor_manager.get_tor",
                        return_value=rx_tm,
                ) as mrx_tm:
                    receive_d = cmd_receive.receive(recv_cfg)
            else:
                KEY_TIMER = 0 if mode == "slow-sender-text" else 99999
                rxw = []
                with mock.patch.object(cmd_receive, "KEY_TIMER", KEY_TIMER):
                    send_d = cmd_send.send(send_cfg)
                    receive_d = cmd_receive.receive(
                        recv_cfg, _debug_stash_wormhole=rxw)
                    # we need to keep KEY_TIMER patched until the receiver
                    # gets far enough to start the timer, which happens after
                    # the code is set
                    if mode == "slow-sender-text":
                        yield rxw[0].get_unverified_key()

            # The sender might fail, leaving the receiver hanging, or vice
            # versa. Make sure we don't wait on one side exclusively
            VERIFY_TIMER = 0 if mode == "slow-text" else 99999
            with mock.patch.object(cmd_receive, "VERIFY_TIMER", VERIFY_TIMER):
                with mock.patch.object(cmd_send, "VERIFY_TIMER", VERIFY_TIMER):
                    if mock_accept or verify:
                        with mock.patch.object(builtins, 'input',
                                return_value='yes') as i:
                            yield gatherResults([send_d, receive_d], True)
                        if verify:
                            s = i.mock_calls[0][1][0]
                            mo = re.search(r'^Verifier (\w+)\. ok\?', s)
                            self.assertTrue(mo, s)
                            sender_verifier = mo.group(1)
                    else:
                        yield gatherResults([send_d, receive_d], True)

            if fake_tor:
                expected_endpoints = [("127.0.0.1", self.rdv_ws_port, False)]
                if mode in ("file", "directory"):
                    expected_endpoints.append(("127.0.0.1", self.transitport, False))
                tx_timing = mtx_tm.call_args[1]["timing"]
                self.assertEqual(tx_tm.endpoints, expected_endpoints)
                self.assertEqual(
                    mtx_tm.mock_calls,
                    [mock.call(reactor, False, None, timing=tx_timing)])
                rx_timing = mrx_tm.call_args[1]["timing"]
                self.assertEqual(rx_tm.endpoints, expected_endpoints)
                self.assertEqual(
                    mrx_tm.mock_calls,
                    [mock.call(reactor, False, None, timing=rx_timing)])

            send_stdout = send_cfg.stdout.getvalue()
            send_stderr = send_cfg.stderr.getvalue()
            receive_stdout = recv_cfg.stdout.getvalue()
            receive_stderr = recv_cfg.stderr.getvalue()

            # all output here comes from a StringIO, which uses \n for
            # newlines, even if we're on windows
            NL = "\n"

        self.maxDiff = None  # show full output for assertion failures

        key_established = ""
        if mode == "slow-text":
            key_established = "Key established, waiting for confirmation...\n"

        self.assertEqual(send_stdout, "")

        # check sender
        if mode == "text" or mode == "slow-text":
            expected = ("Sending text message ({bytes:d} Bytes){NL}"
                        "Wormhole code is: {code}{NL}"
                        "On the other computer, please run:{NL}{NL}"
                        "wormhole receive {verify}{code}{NL}{NL}"
                        "{KE}"
                        "text message sent{NL}").format(
                            bytes=len(message),
                            verify="--verify " if verify else "",
                            code=send_cfg.code,
                            NL=NL,
                            KE=key_established)
            self.failUnlessEqual(send_stderr, expected)
        elif mode == "file":
            self.failUnlessIn(u"Sending {size:s} file named '{name}'{NL}"
                              .format(
                                  size=naturalsize(len(message)),
                                  name=send_filename,
                                  NL=NL), send_stderr)
            self.failUnlessIn(u"Wormhole code is: {code}{NL}"
                              "On the other computer, please run:{NL}{NL}"
                              "wormhole receive {code}{NL}{NL}".format(
                                  code=send_cfg.code, NL=NL), send_stderr)
            self.failUnlessIn(
                u"File sent.. waiting for confirmation{NL}"
                "Confirmation received. Transfer complete.{NL}".format(NL=NL),
                send_stderr)
        elif mode == "directory":
            self.failUnlessIn(u"Sending directory", send_stderr)
            self.failUnlessIn(u"named 'testdir'", send_stderr)
            self.failUnlessIn(u"Wormhole code is: {code}{NL}"
                              "On the other computer, please run:{NL}{NL}"
                              "wormhole receive {code}{NL}{NL}".format(
                                  code=send_cfg.code, NL=NL), send_stderr)
            self.failUnlessIn(
                u"File sent.. waiting for confirmation{NL}"
                "Confirmation received. Transfer complete.{NL}".format(NL=NL),
                send_stderr)

        # check receiver
        if mode in ("text", "slow-text", "slow-sender-text"):
            self.assertEqual(receive_stdout, message + NL)
            if mode == "text":
                if verify:
                    mo = re.search(r'^Verifier (\w+)\.\s*$', receive_stderr)
                    self.assertTrue(mo, receive_stderr)
                    receiver_verifier = mo.group(1)
                    self.assertEqual(sender_verifier, receiver_verifier)
                else:
                    self.assertEqual(receive_stderr, "")
            elif mode == "slow-text":
                self.assertEqual(receive_stderr, key_established)
            elif mode == "slow-sender-text":
                self.assertEqual(receive_stderr, "Waiting for sender...\n")
        elif mode == "file":
            self.failUnlessEqual(receive_stdout, "")
            self.failUnlessIn(u"Receiving file ({size:s}) into: {name!r}".format(
                size=naturalsize(len(message)), name=receive_filename),
                receive_stderr)
            self.failUnlessIn(u"Received file written to ", receive_stderr)
            fn = os.path.join(receive_dir, receive_filename)
            self.failUnless(os.path.exists(fn))
            with open(fn, "r") as f:
                self.failUnlessEqual(f.read(), message)
        elif mode == "directory":
            self.failUnlessEqual(receive_stdout, "")
            want = (r"Receiving directory \(\d+ \w+\) into: {name!r}/"
                    .format(name=receive_dirname))
            self.failUnless(
                re.search(want, receive_stderr), (want, receive_stderr))
            self.failUnlessIn(
                u"Received files written to {name!r}"
                .format(name=receive_dirname),
                receive_stderr)
            fn = os.path.join(receive_dir, receive_dirname)
            self.failUnless(os.path.exists(fn), fn)
            for i in range(5):
                fn = os.path.join(receive_dir, receive_dirname, str(i))
                with open(fn, "r") as f:
                    self.failUnlessEqual(f.read(), message(i))
                self.failUnlessEqual(modes[i], stat.S_IMODE(
                    os.stat(fn).st_mode))

    def test_text(self):
        return self._do_test()

    def test_text_subprocess(self):
        return self._do_test(as_subprocess=True)

    def test_text_tor(self):
        return self._do_test(fake_tor=True)

    def test_text_verify(self):
        return self._do_test(verify=True)

    def test_file(self):
        return self._do_test(mode="file")

    def test_file_override(self):
        return self._do_test(mode="file", override_filename=True)

    def test_file_overwrite(self):
        return self._do_test(mode="file", overwrite=True)

    def test_file_overwrite_mock_accept(self):
        return self._do_test(mode="file", overwrite=True, mock_accept=True)

    def test_file_tor(self):
        return self._do_test(mode="file", fake_tor=True)

    def test_empty_file(self):
        return self._do_test(mode="empty-file")

    def test_directory(self):
        return self._do_test(mode="directory")

    def test_directory_addslash(self):
        return self._do_test(mode="directory", addslash=True)

    def test_directory_override(self):
        return self._do_test(mode="directory", override_filename=True)

    def test_directory_overwrite(self):
        return self._do_test(mode="directory", overwrite=True)

    def test_directory_overwrite_mock_accept(self):
        return self._do_test(
            mode="directory", overwrite=True, mock_accept=True)

    def test_slow_text(self):
        return self._do_test(mode="slow-text")

    def test_slow_sender_text(self):
        return self._do_test(mode="slow-sender-text")

    @inlineCallbacks
    def _do_test_fail(self, mode, failmode):
        assert mode in ("file", "directory")
        assert failmode in ("noclobber", "toobig")
        send_cfg = config("send")
        recv_cfg = config("receive")

        for cfg in [send_cfg, recv_cfg]:
            cfg.hide_progress = True
            cfg.relay_url = self.relayurl
            cfg.transit_helper = ""
            cfg.listen = False
            cfg.code = u"1-abc"
            cfg.stdout = io.StringIO()
            cfg.stderr = io.StringIO()

        send_dir = self.mktemp()
        os.mkdir(send_dir)
        receive_dir = self.mktemp()
        os.mkdir(receive_dir)
        recv_cfg.accept_file = True  # don't ask for permission

        if mode == "file":
            message = "test message\n"
            send_cfg.what = receive_name = send_filename = "testfile"
            fn = os.path.join(send_dir, send_filename)
            with open(fn, "w") as f:
                f.write(message)
            size = os.stat(fn).st_size

        elif mode == "directory":
            # $send_dir/
            # $send_dir/$dirname/
            # $send_dir/$dirname/[12345]
            # cd $send_dir && wormhole send $dirname
            # cd $receive_dir && wormhole receive
            # expect: $receive_dir/$dirname/[12345]

            size = 0
            send_cfg.what = receive_name = send_dirname = "testdir"
            os.mkdir(os.path.join(send_dir, send_dirname))
            for i in range(5):
                path = os.path.join(send_dir, send_dirname, str(i))
                with open(path, "w") as f:
                    f.write("test message %d\n" % i)
                size += os.stat(path).st_size

        if failmode == "noclobber":
            PRESERVE = "don't clobber me\n"
            clobberable = os.path.join(receive_dir, receive_name)
            with open(clobberable, "w") as f:
                f.write(PRESERVE)

        send_cfg.cwd = send_dir
        send_d = cmd_send.send(send_cfg)

        recv_cfg.cwd = receive_dir
        receive_d = cmd_receive.receive(recv_cfg)

        # both sides will fail
        if failmode == "noclobber":
            free_space = 10000000
        else:
            free_space = 0
        with mock.patch(
                "wormhole.cli.cmd_receive.estimate_free_space",
                return_value=free_space):
            f = yield self.assertFailure(send_d, TransferError)
            self.assertEqual(
                str(f), "remote error, transfer abandoned: transfer rejected")
            f = yield self.assertFailure(receive_d, TransferError)
            self.assertEqual(str(f), "transfer rejected")

        send_stdout = send_cfg.stdout.getvalue()
        send_stderr = send_cfg.stderr.getvalue()
        receive_stdout = recv_cfg.stdout.getvalue()
        receive_stderr = recv_cfg.stderr.getvalue()

        # all output here comes from a StringIO, which uses \n for
        # newlines, even if we're on windows
        NL = "\n"

        self.maxDiff = None  # show full output for assertion failures

        self.assertEqual(send_stdout, "")
        self.assertEqual(receive_stdout, "")

        # check sender
        if mode == "file":
            self.failUnlessIn("Sending {size:s} file named '{name}'{NL}"
                              .format(
                                  size=naturalsize(size),
                                  name=send_filename,
                                  NL=NL), send_stderr)
            self.failUnlessIn("Wormhole code is: {code}{NL}"
                              "On the other computer, please run:{NL}{NL}"
                              "wormhole receive {code}{NL}".format(
                                  code=send_cfg.code, NL=NL), send_stderr)
            self.failIfIn(
                "File sent.. waiting for confirmation{NL}"
                "Confirmation received. Transfer complete.{NL}".format(NL=NL),
                send_stderr)
        elif mode == "directory":
            self.failUnlessIn("Sending directory", send_stderr)
            self.failUnlessIn("named 'testdir'", send_stderr)
            self.failUnlessIn("Wormhole code is: {code}{NL}"
                              "On the other computer, please run:{NL}{NL}"
                              "wormhole receive {code}{NL}".format(
                                  code=send_cfg.code, NL=NL), send_stderr)
            self.failIfIn(
                "File sent.. waiting for confirmation{NL}"
                "Confirmation received. Transfer complete.{NL}".format(NL=NL),
                send_stderr)

        # check receiver
        if mode == "file":
            self.failIfIn("Received file written to ", receive_stderr)
            if failmode == "noclobber":
                self.failUnlessIn(
                    "Error: "
                    "refusing to overwrite existing 'testfile'{NL}"
                    .format(NL=NL),
                    receive_stderr)
            else:
                self.failUnlessIn(
                    "Error: "
                    "insufficient free space (0B) for file ({size:d}B){NL}"
                    .format(NL=NL, size=size), receive_stderr)
        elif mode == "directory":
            self.failIfIn(
                "Received files written to {name!r}".format(name=receive_name),
                receive_stderr)
            # want = (r"Receiving directory \(\d+ \w+\) into: {name}/"
            #        .format(name=receive_name))
            # self.failUnless(re.search(want, receive_stderr),
            #                (want, receive_stderr))
            if failmode == "noclobber":
                self.failUnlessIn(
                    "Error: "
                    "refusing to overwrite existing 'testdir'{NL}"
                    .format(NL=NL),
                    receive_stderr)
            else:
                self.failUnlessIn(("Error: "
                                   "insufficient free space (0B) for directory"
                                   " ({size:d}B){NL}").format(
                                       NL=NL, size=size), receive_stderr)

        if failmode == "noclobber":
            fn = os.path.join(receive_dir, receive_name)
            self.failUnless(os.path.exists(fn))
            with open(fn, "r") as f:
                self.failUnlessEqual(f.read(), PRESERVE)

    def test_fail_file_noclobber(self):
        return self._do_test_fail("file", "noclobber")

    def test_fail_directory_noclobber(self):
        return self._do_test_fail("directory", "noclobber")

    def test_fail_file_toobig(self):
        return self._do_test_fail("file", "toobig")

    def test_fail_directory_toobig(self):
        return self._do_test_fail("directory", "toobig")


class ZeroMode(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def test_text(self):
        send_cfg = config("send")
        recv_cfg = config("receive")
        message = "textponies"

        for cfg in [send_cfg, recv_cfg]:
            cfg.hide_progress = True
            cfg.relay_url = self.relayurl
            cfg.transit_helper = ""
            cfg.listen = True
            cfg.zeromode = True
            cfg.stdout = io.StringIO()
            cfg.stderr = io.StringIO()

        send_cfg.text = message

        # send_cfg.cwd = send_dir
        # recv_cfg.cwd = receive_dir

        send_d = cmd_send.send(send_cfg)
        receive_d = cmd_receive.receive(recv_cfg)

        yield gatherResults([send_d, receive_d], True)

        send_stdout = send_cfg.stdout.getvalue()
        send_stderr = send_cfg.stderr.getvalue()
        receive_stdout = recv_cfg.stdout.getvalue()
        receive_stderr = recv_cfg.stderr.getvalue()

        # all output here comes from a StringIO, which uses \n for
        # newlines, even if we're on windows
        NL = "\n"

        self.maxDiff = None  # show full output for assertion failures

        self.assertEqual(send_stdout, "")

        # check sender
        expected = ("Sending text message ({bytes:d} Bytes){NL}"
                    "On the other computer, please run:{NL}"
                    "{NL}"
                    "wormhole receive -0{NL}"
                    "{NL}"
                    "text message sent{NL}").format(
                        bytes=len(message), NL=NL)
        self.failUnlessEqual(send_stderr, expected)

        # check receiver
        self.assertEqual(receive_stdout, message + NL)
        self.assertEqual(receive_stderr, "")


class NotWelcome(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        yield self._setup_relay(error="please upgrade XYZ")
        self.cfg = cfg = config("send")
        cfg.hide_progress = True
        cfg.listen = False
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

    @inlineCallbacks
    def test_sender(self):
        self.cfg.text = "hi"
        self.cfg.code = u"1-abc"

        send_d = cmd_send.send(self.cfg)
        f = yield self.assertFailure(send_d, WelcomeError)
        self.assertEqual(str(f), "please upgrade XYZ")

    @inlineCallbacks
    def test_receiver(self):
        self.cfg.code = u"1-abc"

        receive_d = cmd_receive.receive(self.cfg)
        f = yield self.assertFailure(receive_d, WelcomeError)
        self.assertEqual(str(f), "please upgrade XYZ")


class NoServer(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        yield self._setup_relay(None)
        yield self._relay_server.disownServiceParent()

    @inlineCallbacks
    def test_sender(self):
        cfg = config("send")
        cfg.hide_progress = True
        cfg.listen = False
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

        cfg.text = "hi"
        cfg.code = u"1-abc"

        send_d = cmd_send.send(cfg)
        e = yield self.assertFailure(send_d, ServerConnectionError)
        self.assertIsInstance(e.reason, ConnectionRefusedError)

    @inlineCallbacks
    def test_sender_allocation(self):
        cfg = config("send")
        cfg.hide_progress = True
        cfg.listen = False
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

        cfg.text = "hi"

        send_d = cmd_send.send(cfg)
        e = yield self.assertFailure(send_d, ServerConnectionError)
        self.assertIsInstance(e.reason, ConnectionRefusedError)

    @inlineCallbacks
    def test_receiver(self):
        cfg = config("receive")
        cfg.hide_progress = True
        cfg.listen = False
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

        cfg.code = u"1-abc"

        receive_d = cmd_receive.receive(cfg)
        e = yield self.assertFailure(receive_d, ServerConnectionError)
        self.assertIsInstance(e.reason, ConnectionRefusedError)


class Cleanup(ServerBase, unittest.TestCase):
    def make_config(self):
        cfg = config("send")
        # common options for all tests in this suite
        cfg.hide_progress = True
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()
        return cfg

    @inlineCallbacks
    @mock.patch('sys.stdout')
    def test_text(self, stdout):
        # the rendezvous channel should be deleted after success
        cfg = self.make_config()
        cfg.text = "hello"
        cfg.code = u"1-abc"

        send_d = cmd_send.send(cfg)
        receive_d = cmd_receive.receive(cfg)

        yield send_d
        yield receive_d

        cids = self._rendezvous.get_app(cmd_send.APPID).get_nameplate_ids()
        self.assertEqual(len(cids), 0)

    @inlineCallbacks
    def test_text_wrong_password(self):
        # if the password was wrong, the rendezvous channel should still be
        # deleted
        send_cfg = self.make_config()
        send_cfg.text = "secret message"
        send_cfg.code = u"1-abc"
        send_d = cmd_send.send(send_cfg)

        rx_cfg = self.make_config()
        rx_cfg.code = u"1-WRONG"
        receive_d = cmd_receive.receive(rx_cfg)

        # both sides should be capable of detecting the mismatch
        yield self.assertFailure(send_d, WrongPasswordError)
        yield self.assertFailure(receive_d, WrongPasswordError)

        cids = self._rendezvous.get_app(cmd_send.APPID).get_nameplate_ids()
        self.assertEqual(len(cids), 0)


class ExtractFile(unittest.TestCase):
    def test_filenames(self):
        args = mock.Mock()
        args.relay_url = u""
        ef = cmd_receive.Receiver(args)._extract_file
        extract_dir = os.path.abspath(self.mktemp())

        zf = mock.Mock()
        zi = mock.Mock()
        zi.filename = "ok"
        zi.external_attr = 5 << 16
        expected = os.path.join(extract_dir, "ok")
        with mock.patch.object(cmd_receive.os, "chmod") as chmod:
            ef(zf, zi, extract_dir)
            self.assertEqual(zf.extract.mock_calls,
                             [mock.call(zi.filename, path=extract_dir)])
            self.assertEqual(chmod.mock_calls, [mock.call(expected, 5)])

        zf = mock.Mock()
        zi = mock.Mock()
        zi.filename = "../haha"
        e = self.assertRaises(ValueError, ef, zf, zi, extract_dir)
        self.assertIn("malicious zipfile", str(e))

        zf = mock.Mock()
        zi = mock.Mock()
        zi.filename = "haha//root"  # abspath squashes this, hopefully zipfile
        # does too
        zi.external_attr = 5 << 16
        expected = os.path.join(extract_dir, "haha", "root")
        with mock.patch.object(cmd_receive.os, "chmod") as chmod:
            ef(zf, zi, extract_dir)
            self.assertEqual(zf.extract.mock_calls,
                             [mock.call(zi.filename, path=extract_dir)])
            self.assertEqual(chmod.mock_calls, [mock.call(expected, 5)])

        zf = mock.Mock()
        zi = mock.Mock()
        zi.filename = "/etc/passwd"
        e = self.assertRaises(ValueError, ef, zf, zi, extract_dir)
        self.assertIn("malicious zipfile", str(e))


class AppID(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        yield super(AppID, self).setUp()
        self.cfg = cfg = config("send")
        # common options for all tests in this suite
        cfg.hide_progress = True
        cfg.relay_url = self.relayurl
        cfg.transit_helper = ""
        cfg.stdout = io.StringIO()
        cfg.stderr = io.StringIO()

    @inlineCallbacks
    def test_override(self):
        # make sure we use the overridden appid, not the default
        self.cfg.text = "hello"
        self.cfg.appid = u"appid2"
        self.cfg.code = u"1-abc"

        send_d = cmd_send.send(self.cfg)
        receive_d = cmd_receive.receive(self.cfg)

        yield send_d
        yield receive_d

        used = self._usage_db.execute("SELECT DISTINCT `app_id`"
                                      " FROM `nameplates`").fetchall()
        self.assertEqual(len(used), 1, used)
        self.assertEqual(used[0]["app_id"], u"appid2")


class Welcome(unittest.TestCase):
    def do(self, welcome_message, my_version="2.0"):
        stderr = io.StringIO()
        welcome.handle_welcome(welcome_message, "url", my_version, stderr)
        return stderr.getvalue()

    def test_empty(self):
        stderr = self.do({})
        self.assertEqual(stderr, "")

    def test_version_current(self):
        stderr = self.do({"current_cli_version": "2.0"})
        self.assertEqual(stderr, "")

    def test_version_old(self):
        stderr = self.do({"current_cli_version": "3.0"})
        expected = ("Warning: errors may occur unless both sides are"
                    " running the same version\n"
                    "Server claims 3.0 is current, but ours is 2.0\n")
        self.assertEqual(stderr, expected)

    def test_version_unreleased(self):
        stderr = self.do(
            {
                "current_cli_version": "3.0"
            }, my_version="2.5+middle.something")
        self.assertEqual(stderr, "")

    def test_motd(self):
        stderr = self.do({"motd": "hello"})
        self.assertEqual(stderr, "Server (at url) says:\n hello\n")


class Dispatch(unittest.TestCase):
    @inlineCallbacks
    def test_success(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()
        called = []

        def fake():
            called.append(1)

        yield cli._dispatch_command(reactor, cfg, fake)
        self.assertEqual(called, [1])
        self.assertEqual(cfg.stderr.getvalue(), "")

    @inlineCallbacks
    def test_timing(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()
        cfg.timing = mock.Mock()
        cfg.dump_timing = "filename"

        def fake():
            pass

        yield cli._dispatch_command(reactor, cfg, fake)
        self.assertEqual(cfg.stderr.getvalue(), "")
        self.assertEqual(cfg.timing.mock_calls[-1],
                         mock.call.write("filename", cfg.stderr))

    def test_debug_state_invalid_machine(self):
        cfg = cli.Config()
        with self.assertRaises(UsageError):
            cfg.debug_state = "ZZZ"

    @inlineCallbacks
    def test_debug_state_send(self):
        args = config("send")
        args.debug_state = "B,N,M,S,O,K,SK,R,RC,L,C,T"
        args.stdout = io.StringIO()
        s = cmd_send.Sender(args, reactor)
        d = s.go()
        d.cancel()
        try:
            yield d
        except CancelledError:
            pass
        # just check for at least one state-transition we expected to
        # get logged due to the --debug-state option
        self.assertIn(
            "send.B[S0_empty].close",
            args.stdout.getvalue(),
        )

    @inlineCallbacks
    def test_debug_state_receive(self):
        args = config("receive")
        args.debug_state = "B,N,M,S,O,K,SK,R,RC,L,C,T"
        args.stdout = io.StringIO()
        s = cmd_receive.Receiver(args, reactor)
        d = s.go()
        d.cancel()
        try:
            yield d
        except CancelledError:
            pass
        # just check for at least one state-transition we expected to
        # get logged due to the --debug-state option
        self.assertIn(
            "recv.B[S0_empty].close",
            args.stdout.getvalue(),
        )

    @inlineCallbacks
    def test_wrong_password_error(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()

        def fake():
            raise WrongPasswordError("abcd")

        yield self.assertFailure(
            cli._dispatch_command(reactor, cfg, fake), SystemExit)
        expected = fill("ERROR: " + dedent(WrongPasswordError.__doc__)) + "\n"
        self.assertEqual(cfg.stderr.getvalue(), expected)

    @inlineCallbacks
    def test_welcome_error(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()

        def fake():
            raise WelcomeError("abcd")

        yield self.assertFailure(
            cli._dispatch_command(reactor, cfg, fake), SystemExit)
        expected = (
            fill("ERROR: " + dedent(WelcomeError.__doc__)) + "\n\nabcd\n")
        self.assertEqual(cfg.stderr.getvalue(), expected)

    @inlineCallbacks
    def test_transfer_error(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()

        def fake():
            raise TransferError("abcd")

        yield self.assertFailure(
            cli._dispatch_command(reactor, cfg, fake), SystemExit)
        expected = "TransferError: abcd\n"
        self.assertEqual(cfg.stderr.getvalue(), expected)

    @inlineCallbacks
    def test_server_connection_error(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()

        def fake():
            raise ServerConnectionError("URL", ValueError("abcd"))

        yield self.assertFailure(
            cli._dispatch_command(reactor, cfg, fake), SystemExit)
        expected = fill(
            "ERROR: " + dedent(ServerConnectionError.__doc__)) + "\n"
        expected += "(relay URL was URL)\n"
        expected += "abcd\n"
        self.assertEqual(cfg.stderr.getvalue(), expected)

    @inlineCallbacks
    def test_other_error(self):
        cfg = config("send")
        cfg.stderr = io.StringIO()

        def fake():
            raise ValueError("abcd")

        # I'm seeing unicode problems with the Failure().printTraceback, and
        # the output would be kind of unpredictable anyways, so we'll mock it
        # out here.
        f = mock.Mock()

        def mock_print(file):
            file.write(u"<TRACEBACK>\n")

        f.printTraceback = mock_print
        with mock.patch("wormhole.cli.cli.Failure", return_value=f):
            yield self.assertFailure(
                cli._dispatch_command(reactor, cfg, fake), SystemExit)
        expected = "<TRACEBACK>\nERROR: abcd\n"
        self.assertEqual(cfg.stderr.getvalue(), expected)


class Help(unittest.TestCase):
    def _check_top_level_help(self, got):
        # the main wormhole.cli.cli.wormhole docstring should be in the
        # output, but formatted differently
        self.assertIn("Create a Magic Wormhole and communicate through it.",
                      got)
        self.assertIn("--relay-url", got)
        self.assertIn("Receive a text message, file, or directory", got)

    def test_help(self):
        result = CliRunner().invoke(cli.wormhole, ["help"])
        self._check_top_level_help(result.output)
        self.assertEqual(result.exit_code, 0)

    def test_dash_dash_help(self):
        result = CliRunner().invoke(cli.wormhole, ["--help"])
        self._check_top_level_help(result.output)
        self.assertEqual(result.exit_code, 0)

    def test_inconsistent_receive_code_length(self):
        """
        specifying --code-length without --allocate is an error
        """
        result = CliRunner().invoke(
            cli.wormhole,
            ["receive", "--code-length", "3", "2-foo-bar"]
        )
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("Must use --allocate", result.stdout)

    def test_inconsistent_receive_allocate(self):
        """
        specifying --allocate and a code is an error
        """
        result = CliRunner().invoke(
            cli.wormhole,
            ["receive", "--allocate", "2-foo-bar"]
        )
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("Cannot specify a code", result.stdout)



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.task import Clock
from twisted.internet.interfaces import ITransport
from ...eventual import EventualQueue
from ..._interfaces import IDilationConnector
from ..._dilation.roles import LEADER, FOLLOWER
from ..._dilation.connection import (DilatedConnectionProtocol, encode_record,
                                     KCM, Open, Ack)
from .common import clear_mock_calls


def make_con(role, use_relay=False):
    clock = Clock()
    eq = EventualQueue(clock)
    connector = mock.Mock()
    alsoProvides(connector, IDilationConnector)
    n = mock.Mock()  # pretends to be a Noise object
    n.write_message = mock.Mock(side_effect=[b"handshake"])
    c = DilatedConnectionProtocol(eq, role, "desc", connector, n,
                                  b"outbound_prologue\n", b"inbound_prologue\n")
    if use_relay:
        c.use_relay(b"relay_handshake\n")
    t = mock.Mock()
    alsoProvides(t, ITransport)
    return c, n, connector, t, eq


class Connection(unittest.TestCase):
    def test_hashable(self):
        c, n, connector, t, eq = make_con(LEADER)
        hash(c)

    def test_bad_prologue(self):
        c, n, connector, t, eq = make_con(LEADER)
        c.makeConnection(t)
        d = c.when_disconnected()
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        clear_mock_calls(n, connector, t)

        c.dataReceived(b"prologue\n")
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.loseConnection()])

        eq.flush_sync()
        self.assertNoResult(d)
        c.connectionLost(b"why")
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d), c)

    def _test_no_relay(self, role):
        c, n, connector, t, eq = make_con(role)
        t_kcm = KCM()
        t_open = Open(seqnum=1, scid=0x11223344)
        t_ack = Ack(resp_seqnum=2)
        n.decrypt = mock.Mock(side_effect=[
            encode_record(t_kcm),
            encode_record(t_open),
        ])
        exp_kcm = b"\x00\x00\x00\x03kcm"
        n.encrypt = mock.Mock(side_effect=[b"kcm", b"ack1"])
        m = mock.Mock()  # Manager

        c.makeConnection(t)
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"inbound_prologue\n")

        exp_handshake = b"\x00\x00\x00\x09handshake"
        if role is LEADER:
            # the LEADER sends the Noise handshake message immediately upon
            # receipt of the prologue
            self.assertEqual(n.mock_calls, [mock.call.write_message()])
            self.assertEqual(t.mock_calls, [mock.call.write(exp_handshake)])
        else:
            # however the FOLLOWER waits until receiving the leader's
            # handshake before sending their own
            self.assertEqual(n.mock_calls, [])
            self.assertEqual(t.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])

        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"\x00\x00\x00\x0Ahandshake2")
        if role is LEADER:
            # we're the leader, so we don't send the KCM right away
            self.assertEqual(n.mock_calls, [
                mock.call.read_message(b"handshake2")])
            self.assertEqual(connector.mock_calls, [])
            self.assertEqual(t.mock_calls, [])
            self.assertEqual(c._manager, None)
        else:
            # we're the follower, so we send our Noise handshake, then
            # encrypt and send the KCM immediately
            self.assertEqual(n.mock_calls, [
                mock.call.read_message(b"handshake2"),
                mock.call.write_message(),
                mock.call.encrypt(encode_record(t_kcm)),
            ])
            self.assertEqual(connector.mock_calls, [])
            self.assertEqual(t.mock_calls, [
                mock.call.write(exp_handshake),
                mock.call.write(exp_kcm)])
            self.assertEqual(c._manager, None)
        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"\x00\x00\x00\x03KCM")
        # leader: inbound KCM means we add the candidate
        # follower: inbound KCM means we've been selected.
        # in both cases we notify Connector.add_candidate(), and the Connector
        # decides if/when to call .select()

        self.assertEqual(n.mock_calls, [mock.call.decrypt(b"KCM")])
        self.assertEqual(connector.mock_calls, [mock.call.add_candidate(c)])
        self.assertEqual(t.mock_calls, [])
        clear_mock_calls(n, connector, t, m)

        # now pretend this connection wins (either the Leader decides to use
        # this one among all the candidates, or we're the Follower and the
        # Connector is reacting to add_candidate() by recognizing we're the
        # only candidate there is)
        c.select(m)
        self.assertIdentical(c._manager, m)
        if role is LEADER:
            # TODO: currently Connector.select_and_stop_remaining() is
            # responsible for sending the KCM just before calling c.select()
            # iff we're the LEADER, therefore Connection.select won't send
            # anything. This should be moved to c.select().
            self.assertEqual(n.mock_calls, [])
            self.assertEqual(connector.mock_calls, [])
            self.assertEqual(t.mock_calls, [])
            self.assertEqual(m.mock_calls, [])

            c.send_record(KCM())
            self.assertEqual(n.mock_calls, [
                mock.call.encrypt(encode_record(t_kcm)),
            ])
            self.assertEqual(connector.mock_calls, [])
            self.assertEqual(t.mock_calls, [mock.call.write(exp_kcm)])
            self.assertEqual(m.mock_calls, [])
        else:
            # follower: we already sent the KCM, do nothing
            self.assertEqual(n.mock_calls, [])
            self.assertEqual(connector.mock_calls, [])
            self.assertEqual(t.mock_calls, [])
            self.assertEqual(m.mock_calls, [])
        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"\x00\x00\x00\x04msg1")
        self.assertEqual(n.mock_calls, [mock.call.decrypt(b"msg1")])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [])
        self.assertEqual(m.mock_calls, [mock.call.got_record(t_open)])
        clear_mock_calls(n, connector, t, m)

        c.send_record(t_ack)
        exp_ack = b"\x06\x00\x00\x00\x02"
        self.assertEqual(n.mock_calls, [mock.call.encrypt(exp_ack)])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"\x00\x00\x00\x04ack1")])
        self.assertEqual(m.mock_calls, [])
        clear_mock_calls(n, connector, t, m)

        c.disconnect()
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.loseConnection()])
        self.assertEqual(m.mock_calls, [])
        clear_mock_calls(n, connector, t, m)

    def test_no_relay_leader(self):
        return self._test_no_relay(LEADER)

    def test_no_relay_follower(self):
        return self._test_no_relay(FOLLOWER)

    def test_relay(self):
        c, n, connector, t, eq = make_con(LEADER, use_relay=True)

        c.makeConnection(t)
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"relay_handshake\n")])
        clear_mock_calls(n, connector, t)

        c.dataReceived(b"ok\n")
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        clear_mock_calls(n, connector, t)

        c.dataReceived(b"inbound_prologue\n")
        self.assertEqual(n.mock_calls, [mock.call.write_message()])
        self.assertEqual(connector.mock_calls, [])
        exp_handshake = b"\x00\x00\x00\x09handshake"
        self.assertEqual(t.mock_calls, [mock.call.write(exp_handshake)])
        clear_mock_calls(n, connector, t)

    def test_relay_jilted(self):
        c, n, connector, t, eq = make_con(LEADER, use_relay=True)
        d = c.when_disconnected()

        c.makeConnection(t)
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"relay_handshake\n")])
        clear_mock_calls(n, connector, t)

        c.connectionLost(b"why")
        eq.flush_sync()
        self.assertIdentical(self.successResultOf(d), c)

    def test_relay_bad_response(self):
        c, n, connector, t, eq = make_con(LEADER, use_relay=True)

        c.makeConnection(t)
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"relay_handshake\n")])
        clear_mock_calls(n, connector, t)

        c.dataReceived(b"not ok\n")
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.loseConnection()])
        clear_mock_calls(n, connector, t)

    def test_follower_combined(self):
        c, n, connector, t, eq = make_con(FOLLOWER)
        t_kcm = KCM()
        t_open = Open(seqnum=1, scid=0x11223344)
        n.decrypt = mock.Mock(side_effect=[
            encode_record(t_kcm),
            encode_record(t_open),
        ])
        exp_kcm = b"\x00\x00\x00\x03kcm"
        n.encrypt = mock.Mock(side_effect=[b"kcm", b"ack1"])
        m = mock.Mock()  # Manager

        c.makeConnection(t)
        self.assertEqual(n.mock_calls, [mock.call.start_handshake()])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"inbound_prologue\n")

        exp_handshake = b"\x00\x00\x00\x09handshake"
        # however the FOLLOWER waits until receiving the leader's
        # handshake before sending their own
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(t.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])

        clear_mock_calls(n, connector, t, m)

        c.dataReceived(b"\x00\x00\x00\x0Ahandshake2")
        # we're the follower, so we send our Noise handshake, then
        # encrypt and send the KCM immediately
        self.assertEqual(n.mock_calls, [
            mock.call.read_message(b"handshake2"),
            mock.call.write_message(),
            mock.call.encrypt(encode_record(t_kcm)),
        ])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [
            mock.call.write(exp_handshake),
            mock.call.write(exp_kcm)])
        self.assertEqual(c._manager, None)
        clear_mock_calls(n, connector, t, m)

        # the leader will select a connection, send the KCM, and then
        # immediately send some more data

        kcm_and_msg1 = (b"\x00\x00\x00\x03KCM" +
                        b"\x00\x00\x00\x04msg1")
        c.dataReceived(kcm_and_msg1)

        # follower: inbound KCM means we've been selected.
        # in both cases we notify Connector.add_candidate(), and the Connector
        # decides if/when to call .select()

        self.assertEqual(n.mock_calls, [mock.call.decrypt(b"KCM"),
                                        mock.call.decrypt(b"msg1")])
        self.assertEqual(connector.mock_calls, [mock.call.add_candidate(c)])
        self.assertEqual(t.mock_calls, [])
        clear_mock_calls(n, connector, t, m)

        # now pretend this connection wins (either the Leader decides to use
        # this one among all the candidates, or we're the Follower and the
        # Connector is reacting to add_candidate() by recognizing we're the
        # only candidate there is)
        c.select(m)
        self.assertIdentical(c._manager, m)
        # follower: we already sent the KCM, do nothing
        self.assertEqual(n.mock_calls, [])
        self.assertEqual(connector.mock_calls, [])
        self.assertEqual(t.mock_calls, [])
        self.assertEqual(m.mock_calls, [mock.call.got_record(t_open)])
        clear_mock_calls(n, connector, t, m)



from __future__ import absolute_import, print_function, unicode_literals

from twisted.trial import unittest

from .. import journal
from .._interfaces import IJournal


class Journal(unittest.TestCase):
    def test_journal(self):
        events = []
        j = journal.Journal(lambda: events.append("checkpoint"))
        self.assert_(IJournal.providedBy(j))

        with j.process():
            j.queue_outbound(events.append, "message1")
            j.queue_outbound(events.append, "message2")
            self.assertEqual(events, [])
        self.assertEqual(events, ["checkpoint", "message1", "message2"])

    def test_immediate(self):
        events = []
        j = journal.ImmediateJournal()
        self.assert_(IJournal.providedBy(j))

        with j.process():
            j.queue_outbound(events.append, "message1")
            self.assertEqual(events, ["message1"])
            j.queue_outbound(events.append, "message2")
            self.assertEqual(events, ["message1", "message2"])
        self.assertEqual(events, ["message1", "message2"])
